#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import traceback
from typing import List

from fabric.actor.core.common.constants import Constants, ErrorCodes
from fabric.actor.core.apis.i_mgmt_actor import IMgmtActor
from fabric.actor.core.common.exceptions import ManageException
from fabric.actor.core.manage.kafka.kafka_proxy import KafkaProxy
from fabric.message_bus.messages.close_reservations_avro import CloseReservationsAvro
from fabric.message_bus.messages.delegation_avro import DelegationAvro
from fabric.message_bus.messages.get_delegations_avro import GetDelegationsAvro
from fabric.message_bus.messages.reservation_state_avro import ReservationStateAvro
from fabric.message_bus.messages.result_delegation_avro import ResultDelegationAvro
from fabric.message_bus.messages.result_reservation_avro import ResultReservationAvro
from fabric.message_bus.messages.get_reservations_state_request_avro import GetReservationsStateRequestAvro
from fabric.message_bus.messages.add_slice_avro import AddSliceAvro
from fabric.message_bus.messages.get_reservations_request_avro import GetReservationsRequestAvro
from fabric.message_bus.messages.get_slices_request_avro import GetSlicesRequestAvro
from fabric.message_bus.messages.remove_reservation_avro import RemoveReservationAvro
from fabric.message_bus.messages.remove_slice_avro import RemoveSliceAvro
from fabric.message_bus.messages.reservation_mng import ReservationMng
from fabric.message_bus.messages.result_avro import ResultAvro
from fabric.message_bus.messages.slice_avro import SliceAvro
from fabric.actor.core.util.id import ID
from fabric.message_bus.messages.update_reservation_avro import UpdateReservationAvro
from fabric.message_bus.messages.update_slice_avro import UpdateSliceAvro


class KafkaActor(KafkaProxy, IMgmtActor):
    def get_guid(self) -> ID:
        return self.management_id

    def prepare(self, *, callback_topic:str):
        self.callback_topic = callback_topic

    def get_slices(self, *, id_token: str = None, slice_id: ID = None) -> List[SliceAvro]:
        self.clear_last()

        status = ResultAvro()
        rret_val = None

        try:
            request = GetSlicesRequestAvro()
            request.guid = str(self.management_id)
            request.auth = self.auth
            request.callback_topic = self.callback_topic
            request.message_id = str(ID())
            request.id_token = id_token
            request.slice_id = slice_id

            ret_val = self.producer.produce_sync(topic=self.kafka_topic, record=request)

            self.logger.debug(Constants.management_inter_actor_outbound_message.format(request.name, self.kafka_topic))

            if ret_val:
                message_wrapper = self.message_processor.add_message(message=request)

                with message_wrapper.condition:
                    message_wrapper.condition.wait(Constants.management_api_timeout_in_seconds)

                if not message_wrapper.done:
                    self.logger.debug(Constants.management_api_timeout_occurred)
                    self.message_processor.remove_message(msg_id=request.get_message_id())
                    status.code = ErrorCodes.ErrorTransportTimeout.value
                    status.message = ErrorCodes.ErrorTransportTimeout.name
                else:
                    self.logger.debug(Constants.management_inter_actor_inbound_message.format(message_wrapper.response))
                    status = message_wrapper.response.status
                    if status.code == 0:
                        rret_val = message_wrapper.response.slices

            else:
                self.logger.debug(Constants.management_inter_actor_message_failed.format(
                    request.name, self.kafka_topic))
                status.code = ErrorCodes.ErrorTransportFailure.value
                status.message = ErrorCodes.ErrorTransportFailure.name

        except Exception as e:
            self.last_exception = e
            status.code = ErrorCodes.ErrorInternalError.value
            status.message = ErrorCodes.ErrorInternalError.name
            status.details = traceback.format_exc()

        self.last_status = status

        return rret_val

    def remove_slice(self, *, slice_id: ID) -> bool:
        self.clear_last()
        status = ResultAvro()
        try:
            request = RemoveSliceAvro()
            request.guid = str(self.management_id)
            request.auth = self.auth
            request.callback_topic = self.callback_topic
            request.message_id = str(ID())
            request.slice_id = str(slice_id)
            ret_val = self.producer.produce_sync(topic=self.kafka_topic, record=request)

            self.logger.debug(Constants.management_inter_actor_outbound_message.format(request.name, self.kafka_topic))

            if ret_val:
                message_wrapper = self.message_processor.add_message(message=request)

                with message_wrapper.condition:
                    message_wrapper.condition.wait(Constants.management_api_timeout_in_seconds)

                if not message_wrapper.done:
                    self.logger.debug(Constants.management_api_timeout_occurred)
                    self.message_processor.remove_message(msg_id=request.get_message_id())
                    status.code = ErrorCodes.ErrorTransportTimeout.value
                    status.message = ErrorCodes.ErrorTransportTimeout.name
                else:
                    self.logger.debug(Constants.management_inter_actor_inbound_message.format(message_wrapper.response))
                    status = message_wrapper.response.status
            else:
                self.logger.debug(Constants.management_inter_actor_message_failed.format(
                    request.name, self.kafka_topic))
                status.code = ErrorCodes.ErrorTransportFailure.value
                status.message = ErrorCodes.ErrorTransportFailure.name

        except Exception as e:
            self.last_exception = e
            status.code = ErrorCodes.ErrorInternalError.value
            status.message = ErrorCodes.ErrorInternalError.name
            status.details = traceback.format_exc()

        self.last_status = status

        return status.code == 0

    def add_slice(self, *, slice_obj: SliceAvro) -> ID:
        rret_val = None
        self.clear_last()
        status = ResultAvro()

        try:
            request = AddSliceAvro()
            request.guid = str(self.management_id)
            request.auth = self.auth
            request.callback_topic = self.callback_topic
            request.message_id = str(ID())
            request.slice_obj = slice_obj
            ret_val = self.producer.produce_sync(topic=self.kafka_topic, record=request)

            self.logger.debug(Constants.management_inter_actor_outbound_message.format(request.name, self.kafka_topic))

            if ret_val:
                message_wrapper = self.message_processor.add_message(message=request)

                with message_wrapper.condition:
                    message_wrapper.condition.wait(Constants.management_api_timeout_in_seconds)

                if not message_wrapper.done:
                    self.logger.debug(Constants.management_api_timeout_occurred)
                    self.message_processor.remove_message(msg_id=request.get_message_id())
                    status.code = ErrorCodes.ErrorTransportTimeout.value
                    status.message = ErrorCodes.ErrorTransportTimeout.name
                else:
                    self.logger.debug(Constants.management_inter_actor_inbound_message.format(message_wrapper.response))
                    status = message_wrapper.response.status
                    if status.code == 0:
                        rret_val = ID(uid=message_wrapper.response.get_result())
            else:
                self.logger.debug(Constants.management_inter_actor_message_failed.format(
                    request.name, self.kafka_topic))
                status.code = ErrorCodes.ErrorTransportFailure.value
                status.message = ErrorCodes.ErrorTransportFailure.name

        except Exception as e:
            self.last_exception = e
            status.code = ErrorCodes.ErrorInternalError.value
            status.message = ErrorCodes.ErrorInternalError.name
            status.details = traceback.format_exc()

        self.last_status = status

        return rret_val

    def update_slice(self, *, slice_obj: SliceAvro) -> bool:
        self.clear_last()
        status = ResultAvro()
        try:
            request = UpdateSliceAvro()
            request.guid = str(self.management_id)
            request.auth = self.auth
            request.callback_topic = self.callback_topic
            request.message_id = str(ID())
            request.slice_obj = slice_obj
            ret_val = self.producer.produce_sync(topic=self.kafka_topic, record=request)

            self.logger.debug(Constants.management_inter_actor_outbound_message.format(request.name, self.kafka_topic))

            if ret_val:
                message_wrapper = self.message_processor.add_message(message=request)

                with message_wrapper.condition:
                    message_wrapper.condition.wait(Constants.management_api_timeout_in_seconds)

                if not message_wrapper.done:
                    self.logger.debug(Constants.management_api_timeout_occurred)
                    self.message_processor.remove_message(msg_id=request.get_message_id())
                    status.code = ErrorCodes.ErrorTransportTimeout.value
                    status.message = ErrorCodes.ErrorTransportTimeout.name
                else:
                    self.logger.debug(Constants.management_inter_actor_inbound_message.format(message_wrapper.response))
                    status = message_wrapper.response.status
            else:
                self.logger.debug(Constants.management_inter_actor_message_failed.format(
                    request.name, self.kafka_topic))
                status.code = ErrorCodes.ErrorTransportFailure.value
                status.message = ErrorCodes.ErrorTransportFailure.name

        except Exception as e:
            self.last_exception = e
            status.code = ErrorCodes.ErrorInternalError.value
            status.message = ErrorCodes.ErrorInternalError.name
            status.details = traceback.format_exc()

        self.last_status = status

        return status.code == 0

    def get_reservations(self, *, id_token: str = None, state: int = None,
                         slice_id: ID = None, rid: ID = None) -> List[ReservationMng]:
        self.clear_last()
        response = ResultReservationAvro()
        response.status = ResultAvro()
        try:
            request = GetReservationsRequestAvro()
            request.guid = str(self.management_id)
            request.auth = self.auth
            request.callback_topic = self.callback_topic
            request.message_id = str(ID())
            request.reservation_state = state
            request.id_token = id_token

            if slice_id is not None:
                request.slice_id = str(slice_id)

            if rid is not None:
                request.reservation_id = str(rid)

            ret_val = self.producer.produce_sync(topic=self.kafka_topic, record=request)

            self.logger.debug(Constants.management_inter_actor_outbound_message.format(request.name, self.kafka_topic))
            response.message_id = request.message_id

            if ret_val:
                message_wrapper = self.message_processor.add_message(message=request)

                with message_wrapper.condition:
                    message_wrapper.condition.wait(Constants.management_api_timeout_in_seconds)

                if not message_wrapper.done:
                    self.logger.debug(Constants.management_api_timeout_occurred)
                    self.message_processor.remove_message(msg_id=request.get_message_id())
                    response.status.code = ErrorCodes.ErrorTransportTimeout.value
                    response.status.message = ErrorCodes.ErrorTransportTimeout.name
                else:
                    self.logger.debug(Constants.management_inter_actor_inbound_message.format(message_wrapper.response))
                    response = message_wrapper.response
            else:
                self.logger.debug(Constants.management_inter_actor_message_failed.format(
                    request.name, self.kafka_topic))
                response.status.code = ErrorCodes.ErrorTransportFailure.value
                response.status.message = ErrorCodes.ErrorTransportFailure.name

        except Exception as e:
            self.last_exception = e
            response.status.code = ErrorCodes.ErrorInternalError.value
            response.status.message = ErrorCodes.ErrorInternalError.name
            response.status.details = traceback.format_exc()

        self.last_status = response.status

        return response.reservations

    def get_delegations(self, *, slice_id: ID = None, state: int = None,
                        delegation_id: ID = None, id_token: str = None) -> List[DelegationAvro]:
        self.clear_last()
        response = ResultDelegationAvro()
        response.status = ResultAvro()
        try:
            request = GetDelegationsAvro()
            request.guid = str(self.management_id)
            request.auth = self.auth
            request.callback_topic = self.callback_topic
            request.message_id = str(ID())
            request.delegation_state = state
            request.id_token = id_token

            if slice_id is not None:
                request.slice_id = str(slice_id)

            if delegation_id is not None:
                request.delegation_id = str(delegation_id)

            ret_val = self.producer.produce_sync(topic=self.kafka_topic, record=request)

            self.logger.debug(Constants.management_inter_actor_outbound_message.format(request.name, self.kafka_topic))
            response.message_id = request.message_id

            if ret_val:
                message_wrapper = self.message_processor.add_message(message=request)

                with message_wrapper.condition:
                    message_wrapper.condition.wait(Constants.management_api_timeout_in_seconds)

                if not message_wrapper.done:
                    self.logger.debug(Constants.management_api_timeout_occurred)
                    self.message_processor.remove_message(msg_id=request.get_message_id())
                    response.status.code = ErrorCodes.ErrorTransportTimeout.value
                    response.status.message = ErrorCodes.ErrorTransportTimeout.name
                else:
                    self.logger.debug(Constants.management_inter_actor_inbound_message.format(message_wrapper.response))
                    response = message_wrapper.response
            else:
                self.logger.debug(Constants.management_inter_actor_message_failed.format(
                    request.name, self.kafka_topic))
                response.status.code = ErrorCodes.ErrorTransportFailure.value
                response.status.message = ErrorCodes.ErrorTransportFailure.name

        except Exception as e:
            self.last_exception = e
            response.status.code = ErrorCodes.ErrorInternalError.value
            response.status.message = ErrorCodes.ErrorInternalError.name
            response.status.details = traceback.format_exc()

        self.last_status = response.status

        return response.delegations

    def remove_reservation(self, *, rid: ID) -> bool:
        status = ResultAvro()
        self.clear_last()
        if rid is None:
            self.last_exception = ManageException(Constants.invalid_argument)
            status.set_code(ErrorCodes.ErrorInvalidArguments.value)
            status.set_message(ErrorCodes.ErrorInvalidArguments.name)
            self.last_status = status
            return False

        try:
            request = RemoveReservationAvro()
            request.guid = str(self.management_id)
            request.auth = self.auth
            request.callback_topic = self.callback_topic
            request.message_id = str(ID())
            request.reservation_id = str(rid)

            ret_val = self.producer.produce_sync(topic=self.kafka_topic, record=request)

            self.logger.debug(Constants.management_inter_actor_outbound_message.format(request.name, self.kafka_topic))

            if ret_val:
                message_wrapper = self.message_processor.add_message(message=request)

                with message_wrapper.condition:
                    message_wrapper.condition.wait(Constants.management_api_timeout_in_seconds)

                if not message_wrapper.done:
                    self.logger.debug(Constants.management_api_timeout_occurred)
                    self.message_processor.remove_message(msg_id=request.get_message_id())
                    status.code = ErrorCodes.ErrorTransportTimeout.value
                    status.message = ErrorCodes.ErrorTransportTimeout.name
                else:
                    self.logger.debug(Constants.management_inter_actor_inbound_message.format(message_wrapper.response))
                    status = message_wrapper.response.status
            else:
                self.logger.debug(Constants.management_inter_actor_message_failed.format(
                    request.name, self.kafka_topic))
                status.code = ErrorCodes.ErrorTransportFailure.value
                status.message = ErrorCodes.ErrorTransportFailure.name

        except Exception as e:
            self.last_exception = e
            status.code = ErrorCodes.ErrorInternalError.value
            status.message = ErrorCodes.ErrorInternalError.name
            status.details = traceback.format_exc()

        self.last_status = status

        return status.code == 0

    def close_reservation(self, *, rid: ID) -> bool:
        status = ResultAvro()
        self.clear_last()
        if rid is None:
            self.last_exception = ManageException(Constants.invalid_argument)
            status.set_code(ErrorCodes.ErrorInvalidArguments.value)
            status.set_message(ErrorCodes.ErrorInvalidArguments.name)
            self.last_status = status
            return False

        try:
            request = CloseReservationsAvro()
            request.guid = str(self.management_id)
            request.auth = self.auth
            request.callback_topic = self.callback_topic
            request.message_id = str(ID())
            request.reservation_id = str(rid)

            ret_val = self.producer.produce_sync(topic=self.kafka_topic, record=request)

            self.logger.debug(Constants.management_inter_actor_outbound_message.format(request.name, self.kafka_topic))

            if ret_val:
                message_wrapper = self.message_processor.add_message(message=request)

                with message_wrapper.condition:
                    message_wrapper.condition.wait(Constants.management_api_timeout_in_seconds)

                if not message_wrapper.done:
                    self.logger.debug(Constants.management_api_timeout_occurred)
                    self.message_processor.remove_message(msg_id=request.get_message_id())
                    status.code = ErrorCodes.ErrorTransportTimeout.value
                    status.message = ErrorCodes.ErrorTransportTimeout.name
                else:
                    self.logger.debug(Constants.management_inter_actor_inbound_message.format(message_wrapper.response))
                    status = message_wrapper.response.status
            else:
                self.logger.debug(Constants.management_inter_actor_message_failed.format(
                    request.name, self.kafka_topic))
                status.code = ErrorCodes.ErrorTransportFailure.value
                status.message = ErrorCodes.ErrorTransportFailure.name

        except Exception as e:
            self.last_exception = e
            status.code = ErrorCodes.ErrorInternalError.value
            status.message = ErrorCodes.ErrorInternalError.name
            status.details = traceback.format_exc()

        self.last_status = status

        return status.code == 0

    def close_reservations(self, *, slice_id: ID) -> bool:
        status = ResultAvro()
        self.clear_last()
        if slice_id is None:
            self.last_exception = ManageException(Constants.invalid_argument)
            status.set_code(ErrorCodes.ErrorInvalidArguments.value)
            status.set_message(ErrorCodes.ErrorInvalidArguments.name)
            self.last_status = status
            return False

        try:
            request = CloseReservationsAvro()
            request.guid = str(self.management_id)
            request.auth = self.auth
            request.callback_topic = self.callback_topic
            request.message_id = str(ID())
            request.slice_id = str(slice_id)

            ret_val = self.producer.produce_sync(topic=self.kafka_topic, record=request)

            self.logger.debug(Constants.management_inter_actor_outbound_message.format(request.name, self.kafka_topic))

            if ret_val:
                message_wrapper = self.message_processor.add_message(message=request)

                with message_wrapper.condition:
                    message_wrapper.condition.wait(Constants.management_api_timeout_in_seconds)

                if not message_wrapper.done:
                    self.logger.debug(Constants.management_api_timeout_occurred)
                    self.message_processor.remove_message(msg_id=request.get_message_id())
                    status.code = ErrorCodes.ErrorTransportTimeout.value
                    status.message = ErrorCodes.ErrorTransportTimeout.name
                else:
                    self.logger.debug(Constants.management_inter_actor_inbound_message.format(message_wrapper.response))
                    status = message_wrapper.response.status
            else:
                self.logger.debug(Constants.management_inter_actor_message_failed.format(
                    request.name, self.kafka_topic))
                status.code = ErrorCodes.ErrorTransportFailure.value
                status.message = ErrorCodes.ErrorTransportFailure.name

        except Exception as e:
            self.last_exception = e
            status.code = ErrorCodes.ErrorInternalError.value
            status.message = ErrorCodes.ErrorInternalError.name
            status.details = traceback.format_exc()

        self.last_status = status

        return status.code == 0

    def update_reservation(self, *, reservation: ReservationMng) -> bool:
        status = ResultAvro()
        self.clear_last()
        if reservation is None:
            self.last_exception = ManageException(Constants.invalid_argument)
            status.set_code(ErrorCodes.ErrorInvalidArguments.value)
            status.set_message(ErrorCodes.ErrorInvalidArguments.name)
            self.last_status = status
            return False

        try:
            request = UpdateReservationAvro()
            request.guid = str(self.management_id)
            request.auth = self.auth
            request.callback_topic = self.callback_topic
            request.message_id = str(ID())
            request.reservation = reservation

            ret_val = self.producer.produce_sync(topic=self.kafka_topic, record=request)

            self.logger.debug(Constants.management_inter_actor_outbound_message.format(request.name, self.kafka_topic))

            if ret_val:
                message_wrapper = self.message_processor.add_message(message=request)

                with message_wrapper.condition:
                    message_wrapper.condition.wait(Constants.management_api_timeout_in_seconds)

                if not message_wrapper.done:
                    self.logger.debug(Constants.management_api_timeout_occurred)
                    self.message_processor.remove_message(msg_id=request.get_message_id())
                    status.code = ErrorCodes.ErrorTransportTimeout.value
                    status.message = ErrorCodes.ErrorTransportTimeout.name
                else:
                    self.logger.debug(Constants.management_inter_actor_inbound_message.format(message_wrapper.response))
                    status = message_wrapper.response.status
            else:
                self.logger.debug(Constants.management_inter_actor_message_failed.format(
                    request.name, self.kafka_topic))
                status.code = ErrorCodes.ErrorTransportFailure.value
                status.message = ErrorCodes.ErrorTransportFailure.name

        except Exception as e:
            self.last_exception = e
            status.code = ErrorCodes.ErrorInternalError.value
            status.message = ErrorCodes.ErrorInternalError.name
            status.details = traceback.format_exc()

        self.last_status = status

        return status.get_code() == 0

    def get_reservation_state_for_reservations(self, *, reservation_list: List[ID],
                                               id_token: str = None) -> List[ReservationStateAvro]:
        status = ResultAvro()
        self.clear_last()
        if reservation_list is None:
            self.last_exception = ManageException(Constants.invalid_argument)
            status.set_code(ErrorCodes.ErrorInvalidArguments.value)
            status.set_message(ErrorCodes.ErrorInvalidArguments.name)
            self.last_status = status
            return None

        try:
            request = GetReservationsStateRequestAvro()
            request.guid = str(self.management_id)
            request.auth = self.auth
            request.callback_topic = self.callback_topic
            request.message_id = str(ID())
            request.reservation_ids = []
            request.id_token = id_token
            for r in reservation_list:
                request.reservation_ids.append(str(r))

            ret_val = self.producer.produce_sync(topic=self.kafka_topic, record=request)

            self.logger.debug(Constants.management_inter_actor_outbound_message.format(request.name, self.kafka_topic))

            if ret_val:
                message_wrapper = self.message_processor.add_message(message=request)

                with message_wrapper.condition:
                    message_wrapper.condition.wait(Constants.management_api_timeout_in_seconds)

                if not message_wrapper.done:
                    self.logger.debug(Constants.management_api_timeout_occurred)
                    self.message_processor.remove_message(msg_id=request.get_message_id())
                    status.code = ErrorCodes.ErrorTransportTimeout.value
                    status.message = ErrorCodes.ErrorTransportTimeout.name
                else:
                    self.logger.debug(Constants.management_inter_actor_inbound_message.format(message_wrapper.response))
                    status = message_wrapper.response.status
                    if status.get_code() == 0:
                        return message_wrapper.response.reservation_states
            else:
                self.logger.debug(Constants.management_inter_actor_message_failed.format(
                    request.name, self.kafka_topic))
                status.code = ErrorCodes.ErrorTransportFailure.value
                status.message = ErrorCodes.ErrorTransportFailure.name

        except Exception as e:
            self.last_exception = e
            status.code = ErrorCodes.ErrorInternalError.value
            status.message = ErrorCodes.ErrorInternalError.name
            status.details = traceback.format_exc()

        self.last_status = status

        return None

    def clone(self):
        return KafkaActor(guid=self.management_id,
                          kafka_topic=self.kafka_topic,
                          auth=self.auth, logger=self.logger,
                          message_processor=self.message_processor,
                          producer=self.producer)

    def get_name(self) -> str:
        raise ManageException(Constants.not_implemented)


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import traceback
from datetime import datetime

from fabric.actor.core.apis.i_authority_reservation import IAuthorityReservation
from fabric.actor.core.apis.i_broker_reservation import IBrokerReservation
from fabric.actor.core.apis.i_client_reservation import IClientReservation
from fabric.actor.core.apis.i_reservation import IReservation
from fabric.actor.core.core.authority_policy import AuthorityPolicy
from fabric.actor.core.common.exceptions import AuthorityException
from fabric.actor.core.kernel.resource_set import ResourceSet
from fabric.actor.core.plugins.config.config_token import ConfigToken
from fabric.actor.core.apis.i_resource_control import IResourceControl
from fabric.actor.core.time.term import Term
from fabric.actor.core.time.calendar.authority_calendar import AuthorityCalendar
from fabric.actor.core.util.id import ID
from fabric.actor.core.util.reservation_set import ReservationSet
from fabric.actor.core.util.resource_type import ResourceType


class AuthorityCalendarPolicy(AuthorityPolicy):
    """
    The base for authority policy implementations
    """
    UNSUPPORTED_RESOURCE_TYPE = "Unsupported resource type: {}"

    def __init__(self):
        """
        Creates a new instance.
        """
        super().__init__()
        # If true, we will use lazy revocation.
        self.lazy_close = False
        # Resource control objects indexed by guid. <guid, IResourceControl>
        self.controls_by_guid = {}
        # ResourceControl objects indexed by resource type. <type, IResourceControl>
        self.controls_by_resource_type = {}
        # The authority's calendar. A calendar of all requests
        self.calendar = None
        # Says if the actor has been initialized
        self.initialized = False

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['logger']
        del state['actor']
        del state['clock']
        del state['initialized']

        del state['tickets']

        del state['controls_by_resource_type']
        del state['calendar']

        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.logger = None
        self.actor = None
        self.clock = None
        self.initialized = False

        self.tickets = None

        self.controls_by_resource_type = {}
        self.restore()

    def restore(self):
        """
        Custom restore function. Invoked during recovering the policy object.
        """
        for c in self.controls_by_guid.values():
            try:
                self.register_control_types(control=c)
            except Exception as e:
                raise AuthorityException("Cannot restore resource control e:{}".format(e))

    def initialize(self):
        """
        initialize the policy
        """
        if not self.initialized:
            super().initialize()
            self.calendar = AuthorityCalendar(clock=self.clock)
            self.initialize_controls()
            self.initialized = True

    def initialize_controls(self):
        """
        Initializes all registered controls.
        @raises Exception in case of error
        """
        for control in self.controls_by_guid.values():
            control.set_actor(actor=self.actor)
            control.initialize()

    def donate(self, *, resources: ResourceSet):
        super().donate(resources=resources)
        rc = self.get_control_by_type(rtype=resources.get_type())
        if rc is not None:
            rc.donate(resource_set=resources)
        else:
            raise AuthorityException(self.UNSUPPORTED_RESOURCE_TYPE.format(resources.get_type()))

    def eject(self, *, resources: ResourceSet):
        code = super().unavailable(resources=resources)
        if code == 0:
            rc = self.get_control_by_type(rtype=resources.get_type())
            if rc is not None:
                code = rc.unavailable(resource_set=resources)
            else:
                raise AuthorityException(self.UNSUPPORTED_RESOURCE_TYPE.format(resources.get_type()))
        return code

    def available(self, *, resources: ResourceSet):
        super().available(resources=resources)
        rc = self.get_control_by_type(rtype=resources.get_type())
        if rc is not None:
            rc.available(resource_set=resources)
        else:
            raise AuthorityException(self.UNSUPPORTED_RESOURCE_TYPE.format(resources.get_type()))

    def freed(self, *, resources: ResourceSet):
        super().freed(resources=resources)
        rc = self.get_control_by_type(rtype=resources.get_type())
        if rc is not None:
            rc.freed(resource_set=resources)
        else:
            raise AuthorityException(self.UNSUPPORTED_RESOURCE_TYPE.format(resources.get_type()))

    def release(self, *, resources: ResourceSet):
        super().release(resources=resources)
        rc = self.get_control_by_type(rtype=resources.get_type())
        if rc is not None:
            rc.release(resource_set=resources)
        else:
            raise AuthorityException(self.UNSUPPORTED_RESOURCE_TYPE.format(resources.get_type()))

    def recovery_starting(self):
        super().recovery_starting()
        for c in self.controls_by_guid.values():
            c.recovery_starting()

    def revisit(self, *, reservation: IReservation):
        super().revisit(reservation=reservation)
        if isinstance(reservation, IAuthorityReservation):
            self.calendar.add_closing(reservation=reservation, cycle=self.get_close(term=reservation.get_term()))
            approved = reservation.get_approved_resources()
            if approved is None:
                self.logger.debug("Reservation has no approved resources. Nothing is allocated to it.")

            rtype = approved.get_type()
            self.logger.debug("Resource type for recovered reservation: " + rtype)
            control = self.get_control_by_type(rtype=rtype)
            if control is None:
                raise AuthorityException("Missing resource control")
            control.revisit(reservation=reservation)

    def recovery_ended(self):
        super().recovery_ended()
        for c in self.controls_by_guid.values():
            c.recovery_ended()

    def donate_reservation(self, *, reservation: IClientReservation):
        super().donate_reservation(reservation=reservation)
        rc = self.get_control_by_type(rtype=reservation.get_type())
        if rc is not None:
            rc.donate_reservation(reservation=reservation)
        else:
            raise AuthorityException(self.UNSUPPORTED_RESOURCE_TYPE.format(reservation.get_type()))

    def bind(self, *, reservation: IAuthorityReservation) -> bool:
        if isinstance(reservation, IBrokerReservation):
            return super().bind(reservation=reservation)
        # Simple for now: make sure that this is a valid term and do not modify
        # its start/end time and add it to the calendar. If the request came
        # after its start time, but before its end cycle, add it for allocation
        # to lastAllocatedCycle + 1. If it came after its end cycle, throw.
        current_cycle = self.actor.get_current_cycle()
        approved = reservation.get_requested_term()
        start = self.clock.cycle(when=approved.get_new_start_time())

        if start <= current_cycle:
            end = self.clock.cycle(when=approved.get_end_time())
            if end <= current_cycle:
                self.error(message="The request cannot be redeemed: its term has expired")
            start = current_cycle + 1

        self.calendar.add_request(reservation=reservation, cycle=start)
        close = self.get_close(term=reservation.get_requested_term())
        self.calendar.add_closing(reservation=reservation, cycle=close)
        return False

    def extend(self, *, reservation: IReservation, resources: ResourceSet, term: Term):
        # Simple for now: make sure that this is a valid term and do not modify
        # its start/end time and add it to the calendar. If the request came
        # after its start time, but before its end cycle, add it for allocation
        # to lastAllocatedCycle + 1. If it came after its end cycle, throw an
        # exception.

        if resources is not None and term is not None:
            raise AuthorityException("Not implemented")
        current_cycle = self.actor.get_current_cycle()
        approved = reservation.get_requested_term()
        start = self.clock.cycle(when=approved.get_new_start_time())

        if start <= current_cycle:
            end = self.clock.cycle(when=approved.get_end_time())
            if end <= current_cycle:
                self.error(message="The request cannot be redeemed: its term has expired")
            start = current_cycle + 1

        self.calendar.remove_closing(reservation=reservation)
        self.calendar.add_request(reservation=reservation, cycle=start)
        close = self.get_close(term=reservation.get_requested_term())
        self.calendar.add_closing(reservation=reservation, cycle=close)
        return True

    def correct_deficit(self, *, reservation: IAuthorityReservation):
        if reservation.get_resources() is not None:
            rc = self.get_control_by_type(rtype=reservation.get_resources().get_type())
            if rc is not None:
                self.finish_correct_deficit(rset=rc.correct_deficit(reservation=reservation), reservation=reservation)
            else:
                raise AuthorityException(self.UNSUPPORTED_RESOURCE_TYPE.format(reservation.get_type()))

    def close(self, *, reservation: IReservation):
        self.calendar.remove_schedule_or_in_progress(reservation=reservation)
        if reservation.get_type() is not None:
            rc = self.get_control_by_type(rtype=reservation.get_type())
            if rc is not None:
                rc.close(reservation=reservation)
            else:
                raise AuthorityException(self.UNSUPPORTED_RESOURCE_TYPE.format(reservation.get_type()))

    def closed(self, *, reservation: IReservation):
        if isinstance(reservation, IAuthorityReservation):
            self.calendar.remove_outlay(reservation=reservation)

    def remove(self, *, reservation: IReservation):
        raise AuthorityException("Not implemented")

    def finish(self, *, cycle: int):
        super().finish(cycle=cycle)
        self.calendar.tick(cycle=cycle)

    def assign(self, *, cycle: int):
        try:
            requests = self.get_requests(cycle=cycle)
            self.map_for_cycle(requests=requests, cycle=cycle)
        except Exception as e:
            self.logger.error("error in assign: {}".format(e))

    def map_for_cycle(self, *, requests: ReservationSet, cycle: int):
        """
        Orders mapper request processing for this cycle.
        
        @params requests The requests for this cycle
        @params cycle The cycle
        @raises Exception in case of error
        """
        if requests is None or cycle == 0:
            # self.logger.debug("Authority requests for cycle {} = [none]".format(cycle))
            return

        # self.logger.debug("Authority requests for cycle {} = {}".format(cycle, requests))

        self.map_shrinking(bids=requests)
        self.map_growing(bids=requests)

    def map_shrinking(self, *, bids: ReservationSet):
        """
        Maps reservations that are shrinking or staying the same (extending with
        no flex) in this cycle, and removes them from the bid set.
        
        @param bids set of deferred operations for this cycle (non-null)
        @raises Exception in case of error
        """
        # self.logger.debug("Processing shrinking requests")
        rids_to_remove = []
        for reservation in bids.values():
            adjust = reservation.get_deficit()
            if adjust > 0:
                continue
            if not reservation.is_terminal() and reservation.is_extending_lease():
                if adjust < 0:
                    self.logger.debug("**Shrinking reservation by {}:{}".format(adjust, reservation))
                else:
                    self.logger.debug("**Extending reservation (no flex): {}".format(reservation))
                self.map(reservation=reservation)
                rids_to_remove.append(reservation.get_reservation_id())

        for rid in rids_to_remove:
            bids.remove_by_rid(rid=rid)

    def map_growing(self, *, bids: ReservationSet):
        """
        Maps reservations that are growing in this cycle (redeems or expanding
        extends), and removes them from the bid set.
        
        @param bids set of deferred operations for this cycle (non-null)
        @throws Exception in case of error
        """
        # self.logger.debug("Processing growing requests")
        rids_to_remove = []
        for reservation in bids.values():
            if reservation.is_terminal():
                continue
            adjust = reservation.get_deficit()

            if adjust > 0:
                if reservation.is_extending_lease():
                    self.logger.debug("**Growing reservation by {}:{}".format(adjust, reservation))
                else:
                    self.logger.debug("**Redeeming reservation by {}:{}".format(adjust, reservation))
            self.map(reservation=reservation)
            rids_to_remove.append(reservation.get_reservation_id())

        for rid in rids_to_remove:
            bids.remove_by_rid(rid=rid)

    def map(self, *, reservation: IAuthorityReservation):
        """
        Maps a reservation. Indicates we will approve the request: update its
        expire time in the calendar, and issue a map probe. The map probe will
        result in a retry of the mapper request through bind or
        <code>extend</code> above, which will release the request to the
        associated mapper.
        
        @param reservation: the reservation
        @throws Exception in case of error
        """
        assigned = self.assign_reservation(reservation=reservation)
        if assigned is not None:
            approved = reservation.get_requested_term()
            reservation.set_approved(term=approved, approved_resources=assigned)
            reservation.set_bid_pending(value=False)
        else:
            if not reservation.is_terminal():
                self.logger.debug("Deferring reservation {} for the next cycle: {}".format(
                    reservation, self.actor.get_current_cycle() + 1))
                self.reschedule(reservation=reservation)

    def assign_reservation(self, *, reservation: IAuthorityReservation):
        """
        Assign resources for the given reservation
        
        @params reservation
                   the request
        @returns a set of resources for the request
        @raises Exception in case of error
        """
        rc = self.get_control_by_type(rtype=reservation.get_requested_resources().get_type())
        if rc is not None:
            try:
                return rc.assign(reservation=reservation)
            except Exception as e:
                traceback.print_exc()
                self.logger.error("Could not assign {}".format(e))
                return None
        else:
            raise AuthorityException(self.UNSUPPORTED_RESOURCE_TYPE.format(reservation.get_type()))

    def configuration_complete(self, *, action: str, token: ConfigToken, out_properties: dict):
        super().configuration_complete(action=action, token=token, out_properties=out_properties)
        rc = self.get_control_by_type(rtype=token.get_resource_type())
        if rc is not None:
            rc.configuration_complete(action=action, token=token, out_properties=out_properties)
        else:
            raise AuthorityException(self.UNSUPPORTED_RESOURCE_TYPE.format(token.get_resource_type()))

    def is_expired(self, *, reservation: IReservation) -> bool:
        """
        See if a reservation has expired
        
        @params reservation: reservation
        @return true if the reservation expired; otherwise, return false
        @raises Exception in case of error
        """
        now = datetime.utcnow()
        end = reservation.get_term().get_end_time()

        return now > end

    def reschedule(self, *, reservation: IAuthorityReservation):
        """
        Reschedule a reservation into the calendar
        
        @param reservation the reservation
        """
        self.calendar.remove(reservation=reservation)
        self.calendar.add_request(reservation=reservation, cycle=self.actor.get_current_cycle() + 1)

    def get_close(self, *, term: Term) -> int:
        """
        Return the cycle when a term closes
        
        @params term: the term
        @returns the cycle of the end of a term
        """
        if self.lazy_close:
            return -1
        else:
            return self.clock.cycle(when=term.get_end_time()) + 1

    def get_closing(self, *, cycle: int) -> ReservationSet:
        return self.calendar.get_closing(cycle=cycle)

    def get_requests(self, *, cycle: int) -> ReservationSet:
        return self.calendar.get_requests(cycle=cycle)

    def get_control_by_id(self, *, guid: ID) -> IResourceControl:
        return self.controls_by_guid.get(guid, None)

    def get_control_by_type(self, *, rtype: ResourceType) -> IResourceControl:
        return self.controls_by_resource_type.get(rtype, None)

    def get_control_types(self):
        """
        Returns a reverse map of resource control to resource types. The table is
        indexed by the resource control object and each entry is a linked list of
        resource types.
        
        @returns a table of all of the different control types
        """
        result = {}
        for key, value in self.controls_by_resource_type.items():
            if value not in result:
                result[value] = []
            result[value].append(key)
        return result

    def register_control(self, *, control: IResourceControl):
        """
        Registers the given control for the specified resource type. If the
        policy plugin has already been initialized, the control should be
        initialized.
        
        @param control: the control
        @raises ConfigurationException in case of error
        """
        self.register_control_types(control=control)
        self.controls_by_guid[control.get_guid()] = control

    def register_control_types(self, *, control: IResourceControl):
        types = control.get_types()
        if types is None or len(types) == 0:
            raise AuthorityException("Resource control does not specify any types")
        for t in types:
            if t is None:
                raise AuthorityException("Invalid resource type specified")

        index = 0
        try:
            for rtype in types:
                if rtype in self.controls_by_resource_type:
                    raise AuthorityException("There is already a control associated with resource type {}".format(rtype))
                self.controls_by_resource_type[rtype] = control
                index += 1
        except Exception as e:
            j = 0
            for t in types:
                if t in self.controls_by_resource_type:
                    self.controls_by_resource_type.pop(t)
                    j += 1
                if j == index:
                    break
            raise e



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from fabric.actor.core.apis.i_actor import IActor
from fabric.actor.core.apis.i_authority_reservation import IAuthorityReservation
from fabric.actor.core.apis.i_base_plugin import IBasePlugin
from fabric.actor.core.apis.i_broker_reservation import IBrokerReservation
from fabric.actor.core.apis.i_callback_proxy import ICallbackProxy
from fabric.actor.core.apis.i_controller_callback_proxy import IControllerCallbackProxy
from fabric.actor.core.apis.i_delegation import IDelegation
from fabric.actor.core.apis.i_rpc_request_state import IRPCRequestState
from fabric.actor.core.apis.i_reservation import IReservation
from fabric.actor.core.apis.i_server_reservation import IServerReservation
from fabric.actor.core.common.constants import Constants
from fabric.actor.core.common.exceptions import ProxyException
from fabric.actor.core.delegation.delegation_factory import DelegationFactory
from fabric.actor.core.kernel.client_reservation_factory import ClientReservationFactory
from fabric.actor.core.kernel.controller_reservation_factory import ControllerReservationFactory
from fabric.actor.core.kernel.resource_set import ResourceSet
from fabric.actor.core.proxies.proxy import Proxy
from fabric.actor.core.proxies.local.local_proxy import LocalProxy
from fabric.actor.core.util.resource_data import ResourceData
from fabric.actor.core.util.update_data import UpdateData
from fabric.actor.security.auth_token import AuthToken


class LocalReturn(LocalProxy, IControllerCallbackProxy):
    def __init__(self, *, actor: IActor):
        super().__init__(actor=actor)
        self.callback = True

    def prepare_update_delegation(self, *, delegation: IDelegation, update_data: UpdateData,
                                  callback: ICallbackProxy, caller: AuthToken) -> IRPCRequestState:

        state = LocalProxy.LocalProxyRequestState()
        state.delegation = LocalReturn.pass_delegation(delegation=delegation)
        state.update_data = UpdateData()
        state.update_data.absorb(other=update_data)
        state.callback = callback
        return state

    def _prepare(self, *, reservation: IServerReservation, update_data: UpdateData,
                 callback: ICallbackProxy, caller: AuthToken) -> IRPCRequestState:

        state = LocalProxy.LocalProxyRequestState()
        state.reservation = LocalReturn.pass_reservation(reservation=reservation, plugin=self.get_actor().get_plugin())
        state.update_data = UpdateData()
        state.update_data.absorb(other=update_data)
        state.callback = callback
        return state

    def prepare_update_ticket(self, *, reservation: IBrokerReservation, update_data: UpdateData,
                              callback: ICallbackProxy, caller: AuthToken) -> IRPCRequestState:
        return self._prepare(reservation=reservation, update_data=update_data, callback=callback, caller=caller)

    def prepare_update_lease(self, *, reservation: IAuthorityReservation, update_data, callback: ICallbackProxy,
                             caller: AuthToken) -> IRPCRequestState:
        return self._prepare(reservation=reservation, update_data=update_data, callback=callback, caller=caller)

    @staticmethod
    def pass_reservation(*, reservation: IServerReservation, plugin: IBasePlugin) -> IReservation:
        slice_obj = reservation.get_slice().clone_request()
        term = None

        if reservation.get_term() is None:
            term = reservation.get_requested_term().clone()
        else:
            term = reservation.get_term().clone()

        rset = None

        if reservation.get_resources() is None:
            rset = ResourceSet(units=0, rtype=reservation.get_requested_type(), rdata=ResourceData())
        else:
            rset = LocalReturn.abstract_clone_return(rset=reservation.get_resources())

            concrete = reservation.get_resources().get_resources()

            if concrete is not None:
                cset = None
                try:
                    encoded = concrete.encode(protocol=Constants.protocol_local)
                    cset = Proxy.decode(encoded=encoded, plugin=plugin)
                except Exception as e:
                    raise ProxyException("Error while encoding concrete set {}".format(e))

                if cset is None:
                    raise ProxyException("Unsupported ConcreteSet type: {}".format(type(concrete)))

                rset.set_resources(cset=cset)

        if isinstance(reservation, IBrokerReservation):
            client_reservation = ClientReservationFactory.create(rid=reservation.get_reservation_id(),
                                                                 resources=rset, term=term, slice_object=slice_obj)
            client_reservation.set_ticket_sequence_in(sequence=reservation.get_sequence_out())
            return client_reservation
        else:
            controller_reservation = ControllerReservationFactory.create(rid=reservation.get_reservation_id(),
                                                                         resources=rset, term=term,
                                                                         slice_object=slice_obj)
            controller_reservation.set_lease_sequence_in(sequence=reservation.get_sequence_out())
            return controller_reservation

    @staticmethod
    def pass_delegation(*, delegation: IDelegation) -> IDelegation:
        slice_obj = delegation.get_slice_object().clone_request()

        delegation_new = DelegationFactory.create(did=delegation.get_delegation_id(),
                                                  slice_id=delegation.get_slice_id())
        delegation_new.set_slice_object(slice_object=slice_obj)
        # TODO
        delegation_new.set_graph(delegation.get_graph())
        return delegation_new



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import queue
import threading
import time
from datetime import datetime

from fabric.actor.core.common.constants import Constants
from fabric.actor.core.kernel.reservation_states import ReservationStates
from fabric.actor.core.time.term import Term
from fabric.orchestrator.core.orchestrator_slice import OrchestratorSlice


class SliceDeferThread:
    """
    This runs as a standalone thread started by ContreollerState and deals with slices that have to wait for other
    slices to complete.
    """
    THREAD_SLEEP_TIME = 10000
    DEFAULT_MAX_CREATE_TIME = 600000
    DEFAULT_DELAY_RESOURCE_TYPES = "nlr.vlan ion.vlan ben.vlan"

    def __init__(self):
        self.deferred_slices = []
        self.queue_lock = threading.Lock()
        self.avail = threading.Condition()
        from fabric.actor.core.container.globals import GlobalsSingleton
        self.logger = GlobalsSingleton.get().get_logger()
        self.thread = None

        self.last_slice = None
        self.last_slice_time = None
        self.max_create_wait_time = 0
        self.delay_resource_types = None
        self.stopped = False

        wait_time = GlobalsSingleton.get().get_config().get_runtime_config().get(
            Constants.property_conf_controller_create_wait_time_ms, None)

        if wait_time is None:
            self.max_create_wait_time = self.DEFAULT_MAX_CREATE_TIME
        else:
            self.max_create_wait_time = wait_time

        delay_rtype = GlobalsSingleton.get().get_config().get_runtime_config().get(
            Constants.property_conf_controller_delay_resource_types, None)

        if delay_rtype is None:
            self.delay_resource_types = self.DEFAULT_DELAY_RESOURCE_TYPES.split(" ")
        else:
            self.delay_resource_types = delay_rtype.split(" ")

    def put_tail(self, *, controller_slice: OrchestratorSlice):
        try:
            self.queue_lock.acquire()
            if controller_slice is not None:
                self.deferred_slices.append(controller_slice)
                self.avail.notify_all()

        finally:
            self.queue_lock.release()

    def get_head(self):
        ret_val = None
        try:
            self.queue_lock.acquire()
            self.avail.wait(self.THREAD_SLEEP_TIME)
            if len(self.deferred_slices) > 0:
                ret_val = self.deferred_slices[0]
        except Exception as e:
            ret_val = None
        finally:
            self.queue_lock.release()

        return ret_val

    def remove_head(self):
        try:
            self.queue_lock.acquire()
            self.deferred_slices.pop(0)
        finally:
            self.queue_lock.release()

    def update_last(self, *, controller_slice: OrchestratorSlice):
        if controller_slice is None:
            return

        self.logger.info("Updating last slice with: {}/{}".format(controller_slice.get_slice_urn(),
                                                                  controller_slice.get_slice_id()))
        self.last_slice = controller_slice
        self.last_slice_time = datetime.utcnow()

    def process_slice(self, *, controller_slice: OrchestratorSlice):
        if controller_slice is None:
            return

        if controller_slice != self.last_slice and self.check_computed_reservations(controller_slice=controller_slice) and \
                self.delay_not_done(controller_slice=self.last_slice):
            self.logger.info("Putting slice {}/{} on wait queue".format(controller_slice.get_slice_urn(),
                                                                        controller_slice.get_slice_id()))
            self.put_tail(controller_slice=controller_slice)
        else:
            self.logger.info("Processing slice {}/{} immediately".format(controller_slice.get_slice_urn(),
                                                                         controller_slice.get_slice_id()))
            if self.check_computed_reservations(controller_slice=controller_slice):
                self.update_last(controller_slice=controller_slice)

            self.demand_slice(controller_slice=controller_slice)

    def start(self):
        self.thread = threading.Thread(target=self.run())
        self.thread.setDaemon(True)
        self.thread.setName('SliceDeferThread')
        self.thread.start()

    def stop(self):
        self.stopped = True
        if self.thread is not None:
            self.thread.join()

    def run(self):
        while True:
            controller_slice = self.get_head()
            #self.logger.debug("SliceDeferThread tick")
            time.sleep(1)

            if self.stopped:
                self.logger.debug("SliceDeferThread exiting")
                return

            if controller_slice is None:
                continue

            self.logger.info("Processing previously deferred slice {}/{}".format(self.last_slice.get_slice_urn(),
                                                                                 self.last_slice.get_slice_id()))
            try:
                self.last_slice.lock()
                if self.delay_not_done(controller_slice=self.last_slice):
                    if Term.delta(self.last_slice, datetime.utcnow()) > self.max_create_wait_time:
                        self.logger.info("Maximum wait time exceeded for slice: {}/{}, proceeding anyway".format(
                            self.last_slice.get_slice_urn(), self.last_slice.get_slice_id()))
                    else:
                        continue
            except Exception as e:
                self.logger.error("Exception while checking slice {}/{} e: {}".format(self.last_slice.get_slice_urn(),
                                                                                      self.last_slice.get_slice_id(),
                                                                                      e))
            finally:
                self.last_slice.unlock()

            self.logger.info("Performing demand on deferred slice {}/{}".format(self.last_slice.get_slice_urn(),
                                                                                self.last_slice.get_slice_id()))

            self.update_last(controller_slice=controller_slice)
            try:
                controller_slice.lock()
                self.demand_slice(controller_slice=controller_slice)
            except Exception as e:
                self.logger.error("Exception while demanding slice: {}/{} e: {}".format(self.last_slice.get_slice_urn(),
                                                                                        self.last_slice.get_slice_id(),
                                                                                        e))
            finally:
                self.last_slice.unlock()

            self.remove_head()

    def demand_slice(self, *, controller_slice: OrchestratorSlice):
        if controller_slice is None:
            self.logger.error("demand slice was given a None slice")
            return

        computed_reservations = controller_slice.get_computed_reservations()

        if computed_reservations is None:
            return

        try:
            from fabric.orchestrator.core.orchestrator_state import OrchestratorStateSingleton
            controller = OrchestratorStateSingleton.get().get_management_actor()
            for reservation in computed_reservations:
                self.logger.debug("Issuing demand for reservation: {}".format(reservation.get_reservation_id()))

                if reservation.get_state() != ReservationStates.Unknown:
                    continue

                if not controller.demand(reservation):
                    raise Exception("Could not demand resources: {}".format(controller.get_last_error()))
        except Exception as e:
            self.logger.error("Unable to get orchestrator or demand reservation")
            return

    def check_computed_reservations(self, *, controller_slice: OrchestratorSlice) -> bool:
        if controller_slice is None or controller_slice.get_computed_reservations() is None:
            self.logger.info("Empty slice or no computed reservations")
            return False

        for reservation in controller_slice.get_computed_reservations():
            for drt in self.delay_resource_types:
                if drt == reservation.get_resource_type():
                    self.logger.info("{}/{} has delayed domain".format(controller_slice.get_slice_urn(),
                                                                       controller_slice.get_slice_id()))
                    return True

        self.logger.info("{}/{} has no delayed domains".format(controller_slice.get_slice_urn(),
                                                               controller_slice.get_slice_id()))

        return False

    def delay_not_done(self, *, controller_slice: OrchestratorSlice) -> bool:
        if controller_slice is None:
            return False

        self.logger.info("Checking slice {}/{}".format(controller_slice.get_slice_urn(),
                                                       controller_slice.get_slice_id()))

        controller = None
        all_reservations = None

        try:
            from fabric.orchestrator.core.orchestrator_state import OrchestratorStateSingleton
            controller = OrchestratorStateSingleton.get().get_management_actor()
            all_reservations = controller_slice.get_all_reservations(controller=controller)
        except Exception as e:
            self.logger.error("Exception in delay_not_done for slice: {}/{} e: {}".format(
                controller_slice.get_slice_urn(), controller_slice.get_slice_id(), e))

        if all_reservations is None:
            self.logger.info("Slice: {}/{} has None reservations in delay_not_done".format(
                controller_slice.get_slice_urn(), controller_slice.get_slice_id()))

            return self.check_computed_reservations(controller_slice=controller_slice)
        else:
            if len(all_reservations) <= 0:
                self.logger.info("Slice: {}/{} has empty reservations in delay_not_done".format(
                    controller_slice.get_slice_urn(), controller_slice.get_slice_id()))
                return self.check_computed_reservations(controller_slice=controller_slice)

            for reservation in all_reservations:
                rtype = reservation.get_resource_type()
                for drt in self.delay_resource_types:
                    if drt == rtype:
                        if reservation.get_state() != ReservationStates.Active and \
                            reservation.get_state() != ReservationStates.Closed and \
                            reservation.get_state() != ReservationStates.CloseWait and \
                                reservation.get_state() != ReservationStates.Failed:
                            self.logger.info("Slice: {}/{} has domain {} with reservation: {} that is not yet done".
                                             format(controller_slice.get_slice_urn(), controller_slice.get_slice_id(),
                                                    drt, reservation.get_reservation_id()))
                            return True
            self.logger.info("Slice: {}/{} has no non-final reservations ({})".format(controller_slice.get_slice_urn(),
                                                                                      controller_slice.get_slice_id(),
                                                                                      len(all_reservations)))

            return False



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, List, Dict

from fabric.actor.core.apis.i_delegation import IDelegation
from fabric.actor.core.apis.i_slice import ISlice

if TYPE_CHECKING:
    from fabric.actor.core.apis.i_kernel_reservation import IKernelReservation
    from fabric.actor.core.util.id import ID
    from fabric.actor.core.util.reservation_set import ReservationSet
    from fabric.actor.security.auth_token import AuthToken


class IKernelSlice(ISlice):
    """
    Kernel-level interface for slice objects.
    """
    @abstractmethod
    def set_broker_client(self):
        """
        Marks the slice as a broker client slice (a client slice within
        an authority that represents a broker).
        """

    @abstractmethod
    def set_client(self):
        """
         Marks the slice as a client slice.
        """

    @abstractmethod
    def get_reservations(self) -> ReservationSet:
        """
        Returns the reservation set.

        @return reservation set
        """

    @abstractmethod
    def get_delegations(self) -> Dict[str, IDelegation]:
        """
        Returns the delegations dict.

        @return delegations dict
        """

    @abstractmethod
    def get_reservations_list(self) -> List[IKernelReservation]:
        """
        Returns the reservation set represented as a list. Must be
        called with the kernel lock on to prevent exceptions due to concurrent
        iteration.

        @return a list of reservation
        """

    @abstractmethod
    def is_empty(self) -> bool:
        """
        Checks if the slice is empty.

        @return true if there are no reservations in the slice
        """

    @abstractmethod
    def prepare(self):
        """
        Prepares to register a new slice.  Clears previous state, such
        as list of reservations in the slice.

        @raises Exception if validity checks fail
        """

    @abstractmethod
    def register(self, *, reservation: IKernelReservation):
        """
        Registers a new reservation.

        @param reservation reservation to register

        @throws Exception in case of error
        """

    @abstractmethod
    def soft_lookup(self, *, rid: ID) -> IKernelReservation:
        """
        Looks up a reservation by ID but does not throw error if the
        reservation is not present in the slice.

        @params rid the reservation ID

        @returns the reservation with that ID
        """

    @abstractmethod
    def unregister(self, *, reservation: IKernelReservation):
        """
        Unregisters the reservation from the slice.

        @param reservation reservation to unregister
        """

    @abstractmethod
    def set_owner(self, *, owner: AuthToken):
        """
        Sets the slice owner.

        @param auth the slice owner
        """

    @abstractmethod
    def register_delegation(self, *, delegation: IDelegation):
        """
        Registers a new delegation.

        @param delegation delegation to register

        @throws Exception in case of error
        """

    @abstractmethod
    def soft_lookup_delegation(self, *, did: ID) -> IDelegation:
        """
        Looks up a delegation by ID but does not throw error if the
        delegation is not present in the slice.

        @params did the delegation ID

        @returns the delegation with that ID
        """

    @abstractmethod
    def unregister_delegation(self, *, delegation: IDelegation):
        """
        Unregisters the delegation from the slice.

        @param delegation delegation to unregister
        """



# coding: utf-8

from __future__ import absolute_import
from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from fabric.orchestrator.swagger_server.models.base_model_ import Model
from fabric.orchestrator.swagger_server import util


class Success(Model):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    def __init__(self, value: object=None):  # noqa: E501
        """Success - a model defined in Swagger

        :param value: The value of this Success.  # noqa: E501
        :type value: object
        """
        self.swagger_types = {
            'value': object
        }

        self.attribute_map = {
            'value': 'value'
        }
        self._value = value

    @classmethod
    def from_dict(cls, dikt) -> 'Success':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The Success of this Success.  # noqa: E501
        :rtype: Success
        """
        return util.deserialize_model(dikt, cls)

    @property
    def value(self) -> object:
        """Gets the value of this Success.

        JSON object  # noqa: E501

        :return: The value of this Success.
        :rtype: object
        """
        return self._value

    @value.setter
    def value(self, value: object):
        """Sets the value of this Success.

        JSON object  # noqa: E501

        :param value: The value of this Success.
        :type value: object
        """

        self._value = value



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from fabric.actor.core.common.constants import Constants
from fabric.actor.core.kernel.resource_set import ResourceSet


class PropertiesManager:

    @staticmethod
    def get_request_properties(*, rset: ResourceSet, create: bool) -> dict:
        """
        Get Request properties
        @param rset resource set
        @param create if true, request properties are created if not present
        """
        properties = rset.get_request_properties()
        if properties is None and create:
            properties = {}
            rset.set_request_properties(p=properties)
        return properties

    @staticmethod
    def set_elastic_size(*, rset: ResourceSet, value: bool):
        """
        Marks this resource set to indicate that the request that it represents
        is elastic in size (can use less units than requested)
        @param rset resource set
        @param value value
        """
        properties = PropertiesManager.get_request_properties(rset=rset, create=True)
        properties[Constants.elastic_size] = value
        return properties

    @staticmethod
    def is_elastic_size(*, rset: ResourceSet):
        """
        Checks if the request represented by this resource set is elastic in size
        @param rset resource set
        @return true or false
        """
        result = False
        properties = rset.get_request_properties()
        if properties is not None and Constants.elastic_size in properties:
            result = properties[Constants.elastic_size]
        return result

    @staticmethod
    def set_elastic_time(*, rset: ResourceSet, value: bool):
        """
        Marks this resource set to indicate that the request that it represents
        is elastic in size (can use less units than requested)
        @param rset resource set
        @param value value
        """
        properties = PropertiesManager.get_request_properties(rset=rset, create=True)
        properties[Constants.elastic_time] = value
        rset.set_request_properties(p=properties)
        return rset

    @staticmethod
    def is_elastic_time(*, rset: ResourceSet):
        """
        Checks if the request represented by this resource set is elastic in size
        @param rset resource set
        @return true or false
        """
        result = False
        properties = rset.get_request_properties()
        if properties is not None and Constants.elastic_time in properties:
            result = properties[Constants.elastic_time]
        return result



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from fabric.actor.boot.inventory.resource_pool_factory import ResourcePoolFactory
from fim.graph.abc_property_graph import ABCPropertyGraph
from fim.graph.neo4j_property_graph import Neo4jGraphImporter, Neo4jPropertyGraph
from fim.graph.resources.neo4j_arm import Neo4jARMGraph
from fim.graph.resources.neo4j_cbm import Neo4jCBMGraph


class Neo4jResourcePoolFactory(ResourcePoolFactory):
    """
    Provides methods to load Graph Models and perform various operations on them
    """
    @staticmethod
    def get_neo4j_importer() -> Neo4jGraphImporter:
        """
        get neo4j graph importer
        :return: Neo4jGraphImporter
        """
        from fabric.actor.core.container.globals import GlobalsSingleton
        neo4j_config = GlobalsSingleton.get().get_config().get_global_config().get_neo4j_config()
        logger = GlobalsSingleton.get().get_logger()

        neo4j_graph_importer = Neo4jGraphImporter(url=neo4j_config["url"], user=neo4j_config["user"],
                                                  pswd=neo4j_config["pass"],
                                                  import_host_dir=neo4j_config["import_host_dir"],
                                                  import_dir=neo4j_config["import_dir"], logger=logger)
        return neo4j_graph_importer

    @staticmethod
    def get_arm_graph_from_file(*, filename: str) -> ABCPropertyGraph:
        """
        Load specified file directly with no manipulations or validation
        :param filename:
        :return:
        """
        neo4j_graph_importer = Neo4jResourcePoolFactory.get_neo4j_importer()
        neo4_graph = neo4j_graph_importer.import_graph_from_file(graph_file=filename)
        neo4_graph.validate_graph()

        arm_graph = Neo4jARMGraph(graph=neo4_graph)

        return arm_graph

    @staticmethod
    def get_arm_graph(*, graph_id: str) -> Neo4jARMGraph:
        """
        Load arm graph from neo4j
        :param graph_id: graph_id
        :return: Neo4jARMGraph
        """
        neo4j_graph_importer = Neo4jResourcePoolFactory.get_neo4j_importer()
        arm_graph = Neo4jARMGraph(graph=Neo4jPropertyGraph(graph_id=graph_id, importer=neo4j_graph_importer))

        return arm_graph

    @staticmethod
    def get_neo4j_cbm_empty_graph() -> Neo4jCBMGraph:
        """
        Load cmb empty graph
        :return: Neo4jCBMGraph
        """
        neo4j_graph_importer = Neo4jResourcePoolFactory.get_neo4j_importer()
        combined_broker_model = Neo4jCBMGraph(importer=neo4j_graph_importer, logger=neo4j_graph_importer.log)
        return combined_broker_model

    @staticmethod
    def get_neo4j_cbm_graph_from_database(combined_broker_model_graph_id: str) -> Neo4jCBMGraph:
        """
        Load cbm graph from neo4j
        :param combined_broker_model_graph_id: combined_broker_model_graph_id
        :return: Neo4jCBMGraph
        """
        neo4j_graph_importer = Neo4jResourcePoolFactory.get_neo4j_importer()
        combined_broker_model = Neo4jCBMGraph(graph_id=combined_broker_model_graph_id,
                                              importer=neo4j_graph_importer,
                                              logger=neo4j_graph_importer.log)
        combined_broker_model.validate_graph()
        return combined_broker_model

    @staticmethod
    def get_graph_from_string(*, graph_str: str) -> Neo4jPropertyGraph:
        """
        Load arm graph from neo4j
        :param graph_str: graph_str
        :return: Neo4jPropertyGraph
        """
        neo4j_graph_importer = Neo4jResourcePoolFactory.get_neo4j_importer()
        graph = neo4j_graph_importer.import_graph_from_string_direct(graph_string=graph_str)

        return graph

    @staticmethod
    def delete_graph(*, graph_id: str):
        """
        Delete a graph
        @param graph_id graph id
        """
        neo4j_graph_importer = Neo4jResourcePoolFactory.get_neo4j_importer()
        neo4j_graph_importer.delete_graph(graph_id=graph_id)


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations
from typing import TYPE_CHECKING

from fabric.actor.core.apis.i_reservation import IReservation
from fabric.actor.core.common.constants import Constants
from fabric.actor.core.common.exceptions import PolicyException
from fabric.actor.core.core.unit import Unit, UnitState
from fabric.actor.core.core.unit_set import UnitSet
from fabric.actor.core.policy.free_allocated_set import FreeAllocatedSet
from fabric.actor.core.policy.resource_control import ResourceControl
from fabric.actor.core.util.id import ID
from fabric.actor.core.util.resource_data import ResourceData
from fabric.actor.core.kernel.resource_set import ResourceSet

if TYPE_CHECKING:
    from fabric.actor.core.apis.i_client_reservation import IClientReservation
    from fabric.actor.core.apis.i_authority_reservation import IAuthorityReservation


class LUNControl(ResourceControl):
    def __init__(self):
        super().__init__()
        self.tags = FreeAllocatedSet()
        self.rtype = None

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['authority']
        del state['logger']
        del state['initialized']

        del state['tags']
        del state['rtype']

        return state

    def __setstate__(self, state):
        self.__dict__.update(state)

        self.authority = None
        self.logger = None
        self.initialized = False

        self.tags = FreeAllocatedSet()
        self.rtype = None

    def donate_reservation(self, *, reservation: IClientReservation):
        if self.tags.size() != 0:
            raise PolicyException("only a single source reservation is supported")

        rset = reservation.get_resources()
        rtype = reservation.get_type()
        local = rset.get_local_properties()

        if local is None:
            raise PolicyException(Constants.not_specified_prefix.format("local properties"))

        self.rtype = rtype
        size = 0

        if Constants.property_lun_range_num in local:
            num_range = int(local[Constants.property_lun_range_num])
            for i in range(num_range):
                start_p = Constants.property_start_lun + str(i)
                end_p = Constants.property_end_lun + str(i)
                if start_p in local and end_p in local:
                    start = int(local[start_p])
                    end = int(local[end_p])
                    if start == 0 and end == 0:
                        break
                    for j in range(start, end):
                        self.tags.add_inventory(item=j)

                    size = size + (end - start + 1)
                    self.logger.info("Tag donation: {}:{}-{}:{}".format(self.rtype, start, end, size))

        if size < reservation.get_units():
            raise PolicyException("Insufficient lun tags specified in donated reservation: donated {} rset says: {}".format(size, reservation.get_units()))

    def assign(self, *, reservation: IAuthorityReservation) -> ResourceSet:
        reservation.set_send_with_deficit(value=True)

        if self.tags.size() == 0:
            raise PolicyException("no inventory")

        requested = reservation.get_requested_resources()
        rtype = reservation.get_type()
        current = reservation.get_resources()
        ticket = requested.get_resources()
        rt = ticket.get_ticket()
        ticket_properties = rt.get_properties()

        if current is None:
            needed = ticket.get_units()
            if needed > 1:
                reservation.fail(message="Cannot assign more than 1 LUN per reservation", exception=None)

            if self.tags.get_free() > 0:
                tag = self.tags.allocate()
                rd = ResourceData()
                rd.resource_properties[Constants.unit_lun_tag] = tag
                gained = UnitSet(plugin=self.authority.get_plugin())
                u = Unit(id=ID())
                u.set_resource_type(rtype=rtype)
                u.set_property(name=Constants.unit_lun_tag, value=str(tag))
                if Constants.resource_storage_capacity in ticket_properties:
                    capacity = int(ticket_properties[Constants.resource_storage_capacity])
                    u.set_property(name=Constants.unit_storage_capacity, value=str(capacity))

                gained.add_unit(u=u)
                return ResourceSet(gained=gained, rtype=rtype, rdata=rd)
            else:
                return None
        else:
            return ResourceSet(rtype=rtype)

    def get_tag(self, *, u: Unit):
        tag = None
        if u.get_property(name=Constants.unit_lun_tag) is not None:
            tag = int(u.get_property(name=Constants.unit_lun_tag))
        return tag

    def free(self, *, uset: dict):
        if uset is not None:
            for u in uset:
                try:
                    tag = self.get_tag(u=u)
                    if tag is None:
                        self.logger.error("LUNControl.free(): attempted to free a unit with a missing tag")
                    else:
                        self.logger.debug("LUNControl.free(): freeing tag: {}".format(tag))
                    self.tags.free(item=tag)
                except Exception as e:
                    self.logger.error("Failed to release lun tag {}".format(e))

    def revisit(self, *, reservation: IReservation):
        unit_set = reservation.get_resources().get_resources()
        for u in unit_set.get_set().values:
            try:
                if u.get_state() == UnitState.DEFAULT or \
                        u.get_state() == UnitState.FAILED or \
                        u.get_state() == UnitState.CLOSING or \
                        u.get_state() == UnitState.PRIMING or \
                        u.get_state() == UnitState.ACTIVE or \
                        u.get_state() == UnitState.MODIFYING:
                    tag = self.get_tag(u=u)
                    if tag is None:
                        self.logger.error("LUNControl.revisit(): Recoverying a unit without a tag")
                    else:
                        self.logger.debug("LUNControl.revisit(): reserving tag {} during recovery".format(tag))
                        self.tags.allocate(tag=tag, config_tag=True)

                elif u.get_state() == UnitState.CLOSED:
                    self.logger.debug("LUNControl.revisit(): node is closed. Nothing to recover")
            except Exception as e:
                self.fail(u=u, message="revisit with vmcontrol", e=e)

    def recovery_starting(self):
        self.logger.info("Beginning LUNControl recovery")

    def recovery_ended(self):
        self.logger.info("Completing LUNControl recovery")
        self.logger.debug("Restored LUNControl resource type {} with tags: {}".format(self.rtype, self.tags))



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, List

from fabric.actor.core.apis.i_delegation import IDelegation
from fabric.actor.core.util.id import ID

if TYPE_CHECKING:
    from fabric.actor.core.time.term import Term
    from fabric.actor.core.util.reservation_set import ReservationSet
    from fabric.actor.core.kernel.resource_set import ResourceSet
    from fabric.actor.core.apis.i_reservation import IReservation


class IReservationOperations:
    """
    ReservationOperations</code> defines the core reservations management operations supported by each actor.
    """
    @abstractmethod
    def fail(self, *, rid: ID, message: str):
        """
        Fails the specified reservation.

        Args:
            rid: reservation id
            message: message
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def close_by_rid(self, *, rid: ID):
        """
        Closes the reservation. Note: the reservation must have already been registered with the actor.
        This method may involve either a client or a server side action or both. When called on a broker,
        this method will only close the broker reservation.

        Args:
            rid: reservation id
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def close(self, *, reservation: IReservation):
        """
        Closes the reservation. Note: the reservation must have already been registered with the actor.
        This method may involve either a client or a server side action or both. When called on a broker,
        this method will only close the broker reservation.

        Args:
            reservation: reservation
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def close_reservations(self, *, reservations: ReservationSet):
        """
        Issues a close for every reservation in the set. All exceptions are caught and logged but no exception
        is propagated. No information will be delivered to indicate that some failure has taken place, e.g.,
        failure to communicate with a broker. Inspect the state of individual reservations to
        determine whether/what failures have taken place.

        Args:
            reservations: reservation set
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def extend(self, *, rid: ID, resources: ResourceSet, term: Term):
        """
        Extends the reservation. Note: the reservation must have already been registered with the actor.
        This method may involve either a client or a server side action or both. When called on a broker,
        this method will extend the current ticket and send an update back to the client. When called on a
        site authority, this method will extend the lease and send the new lease back.

        Args:
            reservation: reservation to extend
            rid: reservation id
            resources: resource set describing the resources desired for the extension
            term: term for extension (must extend the current term)
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def get_reservation(self, *, rid: ID) -> IReservation:
        """
        Returns the specified reservation.

        Args:
            rid: reservation id
        Returns:
            reservation
        """

    @abstractmethod
    def get_reservations(self, *, slice_id: ID) -> List[IReservation]:
        """
        Returns all reservations in the given slice.

        Args:
            slice_id: slice id
        Returns:
            reservations
        """

    @abstractmethod
    def register(self, *, reservation: IReservation):
        """
        Registers the reservation with the actor. The reservation must not have been previously
        registered with the actor: there should be no database record for the reservation.

        Args:
            reservation: reservation to register
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def remove_reservation(self, *, reservation: IReservation = None, rid: ID = None):
        """
        Removes the specified reservation. Note: the reservation must have already been registered with the actor.
        This method will unregister the reservation and remove it from the underlying database.
        Only closed and failed reservations can be removed.

        Args:
            reservation: reservation
            rid: reservation id
        Raises:
            Exception if an error occurs or when trying to remove a reservation that is neither failed or closed.
        """

    @abstractmethod
    def re_register(self, *, reservation: IReservation):
        """
        Registers a previously registered reservation with the actor. The reservation must have a database record.

        Args:
            reservation: reservation
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def re_register_delegation(self, *, delegation: IDelegation):
        """
        Registers a previously registered delegation with the actor. The delegation must have a database record.

        Args:
            delegation: delegation
        Raises:
            Exception in case of error
        """

    @abstractmethod
    def unregister(self, *, reservation: IReservation, rid: ID):
        """
        Unregisters the reservation with the actor. The reservation's database record will not be removed.

        Args:
            reservation: reservation
            rid: reservation id
        Raises:
            Exception in case of error
        """



# coding: utf-8

# flake8: noqa
from __future__ import absolute_import
# import models into model package
from fabric.orchestrator.swagger_server.models.success import Success
from fabric.orchestrator.swagger_server.models.version import Version



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations
from typing import TYPE_CHECKING

from fabric.actor.boot.inventory.i_resource_pool_factory import IResourcePoolFactory, ResourcePoolException
from fabric.actor.core.common.constants import Constants
from fabric.actor.core.core.ticket import Ticket
from fabric.actor.core.kernel.client_reservation_factory import ClientReservationFactory
from fabric.actor.core.kernel.resource_set import ResourceSet
from fabric.actor.core.registry.actor_registry import ActorRegistrySingleton
from fabric.actor.core.util.id import ID
from fabric.actor.core.time.term import Term
from fabric.actor.core.util.resource_data import ResourceData

if TYPE_CHECKING:
    from fabric.actor.core.apis.i_client_reservation import IClientReservation
    from fabric.actor.core.apis.i_slice import ISlice
    from fabric.actor.core.delegation.resource_ticket import ResourceTicket
    from fabric.actor.core.apis.i_substrate import ISubstrate
    from fabric.actor.core.common.resource_pool_descriptor import ResourcePoolDescriptor


class ResourcePoolFactory(IResourcePoolFactory):
    def __init__(self):
        # The resource pool descriptor. Its initial version is passed during initialization.
        # The factory can manipulate it as it sees fit and returns it back the the PoolCreator.
        self.desc = None
        # The actor's substrate
        self.substrate = None
        # The authority proxy for this actor.
        self.proxy = None
        # Slice representing the resource pool.
        self.slice_obj = None

    def update_descriptor(self):
        """
        Modifies the resource pool descriptor as needed
        @raises Exception in case of error
        """
        # Use this function to modify the resource pool descriptor, as needed. For example, you can define attributes
        # and resource pool properties needed by the resource pool. Resource pool attributes will become resource
        # properties (of the pool/slice and source reservation), while properties attached to the resource pool
        # descriptor will become local properties.

    def create_term(self) -> Term:
        """
        Creates the term for the source reservation.
        @return Term
        @throws Exception in case of error
        """
        clock = self.substrate.get_actor().get_actor_clock()
        from fabric.actor.core.container.globals import GlobalsSingleton
        now = GlobalsSingleton.get().get_container().get_current_cycle()
        start = self.desc.get_start()
        if start is None:
            start = clock.cycle_start_date(cycle=now)
        end = self.desc.get_end()
        if end is None:
            # export for one year
            length = 1000 * 60 * 60 * 24 * 365
            end = clock.cycle_end_date(cycle=(now + length))

        return Term(start=start, end=end)

    def create_resource_ticket(self, *, term: Term) -> ResourceTicket:
        """
        Creates the resource ticket for the source reservation
        @param term term
        @return ResourceTicket
        @throws Exception in case of error
        """
        try:
            delegation = self.substrate.get_ticket_factory().make_delegation(units=self.desc.get_units(), term=term,
                                                                             rtype=self.desc.get_resource_type())
            ticket = self.substrate.get_ticket_factory().make_ticket(delegation=delegation)
            return ticket
        except Exception as e:
            raise ResourcePoolException("Could not make ticket {}".format(e))

    def create_resource_data(self) -> ResourceData:
        rdata = ResourceData()
        rdata.resource_properties = self.slice_obj.get_resource_properties()
        rdata.local_properties = self.desc.get_pool_properties()
        rdata.resource_properties[Constants.pool_name] = self.slice_obj.get_name()
        return rdata

    def create_source_reservation(self, *, slice_obj: ISlice) -> IClientReservation:
        self.slice_obj = slice_obj
        term = self.create_term()
        resource_ticket = self.create_resource_ticket(term=term)
        ticket = Ticket(ticket=resource_ticket, plugin=self.substrate, authority=self.proxy)
        rdata = self.create_resource_data()
        resources = ResourceSet(concrete=ticket, rtype=self.desc.get_resource_type(), rdata=rdata)
        reservation = ClientReservationFactory.create(rid=ID(), resources=resources, term=term, slice_object=slice_obj)
        ClientReservationFactory.set_as_source(reservation=reservation)
        return reservation

    def get_descriptor(self) -> ResourcePoolDescriptor:
        self.update_descriptor()
        return self.desc

    def set_descriptor(self, *, descriptor: ResourcePoolDescriptor):
        self.desc = descriptor

    def set_substrate(self, *, substrate: ISubstrate):
        self.substrate = substrate
        auth = self.substrate.get_actor().get_identity()
        try:
            self.proxy = ActorRegistrySingleton.get().get_proxy(protocol=Constants.protocol_kafka,
                                                                actor_name=auth.get_name())
            if self.proxy is None:
                raise ResourcePoolException("Missing proxy")
        except Exception as e:
            raise ResourcePoolException("Could not obtain authority proxy: {} {}".format(auth.get_name(), e))



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from fabric.actor.core.apis.i_reservation import ReservationCategory
from fabric.actor.core.kernel.slice import SliceTypes
from fabric.actor.core.manage.kafka.services.kafka_client_actor_service import KafkaClientActorService
from fabric.actor.core.manage.kafka.services.kafka_server_actor_service import KafkaServerActorService
from fabric.message_bus.messages.message import IMessageAvro


class KafkaBrokerService(KafkaClientActorService, KafkaServerActorService):
    def __init__(self):
        super().__init__()

    def process(self, *, message: IMessageAvro):
        callback_topic = message.get_callback_topic()
        result = None

        self.logger.debug("Processing message: {}".format(message.get_message_name()))

        if message.get_message_name() == IMessageAvro.ClaimResources:
            result = self.claim(request=message)
        elif message.get_message_name() == IMessageAvro.ReclaimResources:
            result = self.reclaim(request=message)
        elif message.get_message_name() == IMessageAvro.AddReservation:
            result = self.add_reservation(request=message)
        elif message.get_message_name() == IMessageAvro.AddReservations:
            result = self.add_reservations(request=message)
        elif message.get_message_name() == IMessageAvro.DemandReservation:
            result = self.demand_reservation(request=message)
        elif message.get_message_name() == IMessageAvro.GetActorsRequest:
            result = self.get_brokers(request=message)
        elif message.get_message_name() == IMessageAvro.GetPoolInfoRequest:
            result = self.get_pool_info(request=message)
        elif message.get_message_name() == IMessageAvro.ExtendReservation:
            result = self.extend_reservation(request=message)
        elif message.get_message_name() == IMessageAvro.GetReservationsRequest and \
                message.get_reservation_type() is not None and \
                message.get_reservation_type() == ReservationCategory.Broker.name:
            result = self.get_broker_reservations(request=message)
        elif message.get_message_name() == IMessageAvro.GetSlicesRequest and \
                message.get_slice_type() is not None and \
                message.get_slice_type() == SliceTypes.InventorySlice.name:
            result = self.get_inventory_slices(request=message)
        elif message.get_message_name() == IMessageAvro.GetReservationsRequest and \
                message.get_reservation_type() is not None and \
                message.get_reservation_type() == ReservationCategory.Client.name:
            result = self.get_inventory_reservations(request=message)
        elif message.get_message_name() == IMessageAvro.GetSlicesRequest and \
                message.get_slice_type() is not None and \
                message.get_slice_type() == SliceTypes.ClientSlice.name:
            result = self.get_client_slices(request=message)
        elif message.get_message_name() == IMessageAvro.AddSlice and message.slice_obj is not None and \
                (message.slice_obj.is_client_slice() or message.slice_obj.is_broker_client_slice()):
            result = self.add_client_slice(request=message)
        else:
            super().process(message=message)
            return

        if callback_topic is None:
            self.logger.debug("No callback specified, ignoring the message")

        if result is None:
            self.logger.error("No response generated {}".format(result))
            return

        if self.producer.produce_sync(topic=callback_topic, record=result):
            self.logger.debug("Successfully send back response: {}".format(result.to_dict()))
        else:
            self.logger.debug("Failed to send back response: {}".format(result.to_dict()))




#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from fabric.actor.core.apis.i_actor import IActor
    from fabric.actor.core.common.resource_vector import ResourceVector
    from fabric.actor.core.delegation.resource_delegation import ResourceDelegation
    from fabric.actor.core.delegation.resource_ticket import ResourceTicket
    from fabric.actor.core.time.term import Term
    from fabric.actor.core.util.id import ID
    from fabric.actor.core.util.resource_type import ResourceType


class IResourceTicketFactory:
    """
    Interface for Factory class for Resource Tickets
    """
    @abstractmethod
    def set_actor(self, *, actor: IActor):
        """
        Sets the actor this factory belongs to.
        @param actor actor
        """

    @abstractmethod
    def get_actor(self) -> IActor:
        """
        Returns the actor represented by this factory.
        @return actor
        """

    @abstractmethod
    def initialize(self):
        """
        Initializes the factory.
        @throws Exception in case of error
        """

    @abstractmethod
    def make_delegation(self, *, units: int, vector: ResourceVector, term: Term, rtype: ResourceType,
                        sources: list = None, bins: list = None, properties: dict = None,
                        holder: ID = None) -> ResourceDelegation:
        """
        Creates a new ResourceDelegation
        @param units number of units
        @param vector resource vector
        @param term term
        @param rtype resource type
        @param sources source bins used for the delegation
        @param bins bins references by the delegation
        @param properties properties list
        @param holder identifier of the holder of the delegation
        @return ResourceDelegation
        @throws Exception in case of error
        """

    @abstractmethod
    def make_ticket(self, *, source: ResourceTicket, delegation: ResourceDelegation) -> ResourceTicket:
        """
        Creates a new ticket from the specified source tickets.
        Note: all sources must share the same root delegation.
        @param source source
        @param delegation delegation
        @return ResourceTicket
        @throws TicketException in case of error
        """

    @abstractmethod
    def clone(self, *, original: ResourceTicket) -> ResourceTicket:
        """
        Makes a deep clone of the specified resource ticket.
        @param original original
        @return ResourceTicket
        """

    @abstractmethod
    def to_json(self, *, ticket: ResourceTicket) -> str:
        """
        Converts the resource ticket to JSON.
        @param ticket ticket to convert
        @return JSON string representation
        """

    @abstractmethod
    def from_json(self, *, incoming: dict) -> ResourceTicket:
        """
        Obtains a resource ticket from JSON.
        @param incoming JSON string representation
        @return resource ticket
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import threading
import traceback

from fabric.actor.core.common.exceptions import ManageException
from fabric.message_bus.consumer import AvroConsumerApi
from fabric.message_bus.messages.message import IMessageAvro


class MessageWrapper:
    def __init__(self, *, message: IMessageAvro):
        self.message = message
        self.condition = threading.Condition()
        self.done = False
        self.response = None


class KafkaMgmtMessageProcessor(AvroConsumerApi):
    def __init__(self, *, conf: dict, key_schema, record_schema, topics, batch_size=5, logger=None):
        super().__init__(conf=conf, key_schema=key_schema, record_schema=record_schema, topics=topics,
                         batch_size=batch_size, logger=logger)
        self.thread_lock = threading.Lock()
        self.thread = None
        self.messages = {}
        self.lock = threading.Lock()
        self.logger = logger

    def start(self):
        try:
            self.thread_lock.acquire()
            if self.thread is not None:
                raise ManageException("KafkaMgmtMessageProcessor has already been started")

            self.thread = threading.Thread(target=self.consume_auto)
            self.thread.setName("KafkaMgmtMessageProcessor")
            self.thread.setDaemon(True)
            self.thread.start()
            self.logger.debug("KafkaMgmtMessageProcessor has been started")
        finally:
            self.thread_lock.release()

    def stop(self):
        self.shutdown()
        try:
            self.thread_lock.acquire()
            temp = self.thread
            self.thread = None
            if temp is not None:
                self.logger.warning("It seems that the KafkaMgmtMessageProcessor thread is running. Interrupting it")
                try:
                    temp.join()
                except Exception as e:
                    self.logger.error("Could not join KafkaMgmtMessageProcessor thread {}".format(e))
                finally:
                    self.thread_lock.release()
        finally:
            if self.thread_lock is not None and self.thread_lock.locked():
                self.thread_lock.release()

    def handle_message(self, *, message: IMessageAvro):
        try:
            message_id = message.get_message_id()

            request = self.remove_message(msg_id=message_id)
            if request is None:
                self.logger.error("No corresponding request found for message_id: {}".format(message_id))
                self.logger.error("Discarding the message: {}".format(message))
                return
            with request.condition:
                request.done = True
                request.response = message
                request.condition.notify_all()

        except Exception as e:
            traceback.print_exc()
            self.logger.error(e)
            self.logger.error("Discarding the incoming message {}".format(message))

    def add_message(self, *, message: IMessageAvro) -> MessageWrapper:
        result = None
        try:
            msg_id = message.get_message_id()
            if msg_id is not None:
                self.thread_lock.acquire()
                result = MessageWrapper(message=message)
                if self.messages.get(msg_id, None) is not None:
                    print("Discarding the message, message with id: {} already exists".format(msg_id))

                self.messages[msg_id] = result
        finally:
            self.thread_lock.release()
        return result

    def remove_message(self, *, msg_id: str):
        try:
            self.thread_lock.acquire()
            if self.messages.get(msg_id, None) is not None:
                message = self.messages.pop(msg_id)
                return message
        finally:
            self.thread_lock.release()
        return None



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import threading
from typing import TYPE_CHECKING

from fabric.actor.boot.inventory.neo4j_resource_pool_factory import Neo4jResourcePoolFactory
from fabric.actor.core.apis.i_delegation import IDelegation
from fabric.actor.core.common.constants import Constants
from fabric.actor.core.common.exceptions import BrokerException
from fabric.actor.core.policy.broker_calendar_policy import BrokerCalendarPolicy
from fabric.actor.core.time.actor_clock import ActorClock
from fabric.actor.core.util.bids import Bids
from fabric.actor.core.util.prop_list import PropList
from fabric.actor.core.util.reservation_set import ReservationSet

if TYPE_CHECKING:
    from fabric.actor.core.apis.i_broker_reservation import IBrokerReservation
    from fabric.actor.core.apis.i_client_reservation import IClientReservation
    from fabric.actor.core.apis.i_broker import IBroker
    from fabric.actor.core.time.term import Term


class BrokerSimplePolicy(BrokerCalendarPolicy):
    """
    BrokerSimplePolicy is a simple implementation of the broker policy interface. It buffers requests for allocation
    periods and when it performs allocations of resources it does so in FIFO order giving preference to extending
    requests.
    """
    REQUEST_TYPE = "requestType"
    PropertyAllocationHorizon = "allocation.horizon"

    # The amount of time over specific policy decisions the broker
    # must add when communicating with other brokers as a client (e.g.
    # renew()). Clock skew must be at least one if the broker is ticked after
    # the upstream broker(s). At some point in time we may want this to not
    # be static and learn it from what we see in the system, but for now it
    # is static.
    CLOCK_SKEW = 1

    # Number of cycles between two consecutive allocations.
    CALL_INTERVAL = 1

    #  How far in the future is the broker allocating resources
    ADVANCE_TIME = 3

    def __init__(self, *, actor: IBroker):
        super().__init__(actor=actor)
        self.last_allocation = -1
        self.allocation_horizon = 0
        self.ready = False

        self.delegations = {}
        self.combined_broker_model = None
        self.combined_broker_model_graph_id = None

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['logger']
        del state['actor']
        del state['clock']
        del state['initialized']

        del state['delegations']
        del state['combined_broker_model']
        del state['for_approval']
        del state['lock']

        del state['calendar']

        del state['last_allocation']
        del state['allocation_horizon']
        del state['ready']

        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.logger = None
        self.actor = None
        self.clock = None
        self.initialized = False

        self.delegations = {}
        self.combined_broker_model = None
        self.load_combined_broker_model()
        self.lock = threading.Lock()
        self.calendar = None

        if self.required_approval:
            self.for_approval = ReservationSet()

        self.last_allocation = -1
        self.allocation_horizon = 0
        self.ready = False

    def load_combined_broker_model(self):
        if self.combined_broker_model_graph_id is None:
            self.logger.debug("Creating an empty Combined Broker Model Graph")

            self.combined_broker_model = Neo4jResourcePoolFactory.get_neo4j_cbm_empty_graph()
            self.combined_broker_model_graph_id = self.combined_broker_model.get_graph_id()

            self.logger.debug("Empty Combined Broker Model Graph created: {}".format(
                self.combined_broker_model_graph_id))
        else:
            self.logger.debug("Loading an existing Combined Broker Model Graph: {}".format(
                self.combined_broker_model_graph_id))

            self.combined_broker_model = Neo4jResourcePoolFactory.get_neo4j_cbm_graph_from_database(
                combined_broker_model_graph_id=self.combined_broker_model_graph_id)
            self.logger.debug(
                "Successfully loaded an existing Combined Broker Model Graph: {}".format(
                    self.combined_broker_model_graph_id))

    def initialize(self):
        if not self.initialized:
            super().initialize()
            self.load_combined_broker_model()
            self.initialized = True

    def configure(self, *, properties: dict):
        if self.PropertyAllocationHorizon in properties:
            self.allocation_horizon = int(properties[self.PropertyAllocationHorizon])
        else:
            self.logger.warning("Invalid property value key: {}".format(self.PropertyAllocationHorizon))

    def create_source_dict(self, *, sources: list) -> dict:
        result = {}
        for s in sources:
            rtype = s.get_type()
            if rtype in sources:
                self.logger.debug("sourcehash already contains key - only allow one source per type")
            else:
                result[rtype] = s
        return result

    def switch_source(self, *, sources: list, start_time: int, source_hash: dict):
        for s in sources:
            extending_for_source = self.calendar.get_request(source=s, cycle=start_time)
            for r in extending_for_source.values():
                if r.get_requested_type() != s.get_type():
                    # this reservation has changed its type
                    if r.get_requested_type() in source_hash:
                        other_source = source_hash[r.get_requested_type()]
                        self.calendar.remove_request(reservation=r, source=s)
                        self.calendar.add_request(reservation=r, cycle=start_time, source=other_source)
                    else:
                        r.fail_warn("This agent has no resources to satisfy a request for type: {}".format(
                            r.get_requested_type()))

    def allocate_extending(self, *, sources: list, start_time: int):
        """
        Iterates through all of the sources and allocates resources to
        their extending bids based on FIFO. Only allocates requests with types
        specified in requestTypes or all requests if requestTypes is None

        @param sources sources for this allocation
        @param start_time when the allocation begins

        @throws Exception in case of error
        """
        for s in sources:
            extending_for_source = self.calendar.get_request(source=s, cycle=start_time)
            if extending_for_source is None or extending_for_source.size() == 0:
                self.logger.debug("There are no extends for source: {}".format(s))
            else:
                self.logger.debug("There are {} extend reservations".format(extending_for_source.size()))

                for extend in extending_for_source.values():
                    self.logger.info("Extend res {}".format(extend))

                    wanted = extend.get_requested_units()
                    self.logger.debug("ASPPlugin:allocateBids: allocating source rid({}) to extendDynamicRes({}) "
                                      "that has props:{}".format(s.get_reservation_id(), extend.get_reservation_id(),
                                                                 extend.get_requested_resources().
                                                                 get_request_properties()))
                    satisfy = self.verify_wanted_resources(reservation=extend, source=s, start_time=start_time,
                                                           wanted=wanted)

                    if satisfy:
                        self.satisfy_allocation(reservation=extend, source=s, resource_share=wanted,
                                                start_res_time=start_time)
                    extending_for_source.remove(extend)

    def allocate_new_bids(self, *, new_bids: list, source_hash: dict, start_time: int):
        """
        Iterates through all of the new bids and allocates resources
        based on FIFO. Only allocates requests with types specified in
        requestTypes or all requests if requestTypes
        is null. Requests are associated with their appropriate source.

        @param new_bids new bids to be allocated
        @param source_hash the sources for this allocation
        @param start_time when the allocation begins

        @throws Exception in case of error
        """
        for reservation in new_bids:
            if not reservation.is_extending_ticket():
                if reservation.get_requested_type() in source_hash:
                    source = source_hash[reservation.get_requested_type()]
                    reservation.set_source(source)
                    wanted = reservation.get_requested_units()
                    satisfy = self.verify_wanted_resources(reservation=reservation, source=source,
                                                           start_time=start_time,
                                                           wanted=wanted)
                    if satisfy:
                        self.satisfy_allocation(reservation=reservation, source=source, resource_share=wanted,
                                                start_res_time=start_time)
                else:
                    reservation.fail("There are no sources of type {}".format(reservation.get_requested_type()))

            new_bids.remove(reservation)

    def allocate(self, *, cycle: int):
        """
        Runs an allocation of requests against all sources, given a set
        of pending bids by using FCFS giving extends priority. This simple
        policy runs through the bids in iterator order, and gives every request
        min(requested,available) until we run out. It gives
        priority to extending requests compared to new requests.<p>The
        implementation assumes that there is only 1 source reservation per
        resource type. It also makes the assumption that we allocate resources
        sequentially in time, so there are never less resources available in
        the future.
        @param cycle cycle
        """
        # This method should first decide whether to run an allocation on the
        # current cycle. If the answer is yes, then it has to decide on the
        # start time of reservations it is going to allocate.
        if self.get_next_allocation(cycle=cycle) != cycle:
            self.logger.debug("Next allocation cycle {} != {}".format(self.get_next_allocation(cycle=cycle), cycle))
            return

        self.last_allocation = cycle
        # Determine the cycle for which the agent is allocating resources
        start_time = self.get_start_for_allocation(allocation_cycle=cycle)

        all_bids = self.calendar.get_requests(cycle=start_time)

        # If there are no extending and no new requests - return
        if all_bids is None or all_bids.size() == 0:
            self.logger.debug("No requests for allocation start cycle {}".format(start_time))
            return

        self.logger.debug("There are {} bids for cycle {}".format(all_bids.size(), start_time))

        # Put source on a Hashtable for the new request allocation
        holdings = self.calendar.get_holdings(self.clock.cycle(cycle=start_time))
        sources = holdings.values()
        source_hash = self.create_source_dict(sources=sources)

        self.switch_source(sources=sources, start_time=start_time, source_hash=source_hash)

        self.allocate_extending(sources=sources, start_time=start_time)

        self.allocate_new_bids(new_bids=all_bids.values(), source_hash=source_hash, start_time=start_time)

    def bind(self, *, reservation: IBrokerReservation) -> bool:
        term = reservation.get_requested_term()
        self.logger.info("SlottedAgent bind arrived at cycle {} requested term {}".format(
            self.actor.get_current_cycle(), term))

        bid_cycle = self.get_allocation(reservation=reservation)

        self.calendar.add_request(reservation=reservation, cycle=bid_cycle)

        return False

    def bind_delegation(self, *, delegation: IDelegation) -> bool:
        try:
            self.lock.acquire()
            self.delegations[delegation.get_delegation_id()] = delegation
        finally:
            self.lock.release()

        return False

    def calculate_available(self, *, start_time: int, source: IClientReservation) -> int:
        """
        Calculates the available resources from the specified source for
        the specified start time.

        @param start_time start of proposed ticket
        @param source source whose resources are being counted

        @return number of available units
        """
        start_date = self.clock.cycle_start_date(cycle=start_time)

        outlays = self.calendar.get_outlays(source=source, time=start_date)
        c = self.count(rvset=outlays, when=start_date)
        active = c.count_active(resource_type=source.get_type())

        available = source.get_units(when=start_date) - active

        self.logger.debug("There are {} resources available - extend".format(available))
        self.logger.debug("There are {} resources active".format(active))
        return available

    def extend_broker(self, *, reservation: IBrokerReservation) -> bool:
        requested_term = reservation.get_requested_term()
        self.logger.info("SlottedAgent extend arrived at cycle {} requested term {}".format(
            self.actor.get_current_cycle(), requested_term))

        source = reservation.get_source()

        if source is None:
            self.error(message="cannot find parent ticket for extend")

        if source.is_failed():
            self.error(message="parent ticket could not be renewed")

        bid_cycle = self.get_allocation(reservation=reservation)

        self.calendar.add_request(reservation=reservation, cycle=bid_cycle, source=source)
        self.calendar.add_request(reservation=reservation, cycle=bid_cycle)

        return False

    def extract_ticket(self, *, reservation: IBrokerReservation, source: IClientReservation,
                       approved: Term, resource_share: int):
        mine = None
        ticket = source.get_resources()
        try:
            if not source.get_term().contains(term=approved):
                reservation.fail(message="Source term does not contain requested term or reservation is not elastic in "
                                         "time: sourceterm={} resterm={}".format(source.get_term(), approved),
                                 exception=None)
            else:
                delegation = self.actor.get_plugin().get_ticket_factory().make_delegation(
                    units=resource_share, term=approved, rtype=ticket.get_type(),
                    holder=self.get_client_id(reservation=reservation))
                mine = self.extract(source=ticket, delegation=delegation)

                req_properties = reservation.get_requested_resources().get_request_properties()
                mine.set_request_properties(p=req_properties)
        except Exception as e:
            self.logger.error("Term not satisfied: Mapper extract failed: has: {} {}".format(
                ticket.get_concrete_units(), e))
            raise BrokerException("Term not satisfied: Mapper extract failed: has: {} {}".format(
                ticket.get_concrete_units(), e))

        if mine is not None and not reservation.is_failed():
            reservation.set_approved(term=approved, approved_resources=mine)
            reservation.set_source(source=source)

            self.logger.debug("allocated: {} for term {}".format(mine.get_units(), approved))
            self.logger.debug("resource_share {} mine {}".format(resource_share, mine.get_units()))

        return mine

    def formulate_bids(self, *, cycle: int) -> Bids:
        pending = self.calendar.get_pending()
        renewing = self.calendar.get_renewing(cycle=cycle)
        extending = self.process_renewing(renewing=renewing, pending=pending)
        return Bids(ticketing=ReservationSet(), extending=extending)

    def get_allocation(self, *, reservation: IBrokerReservation) -> int:
        if not self.ready:
            self.error(message="Agent not ready to accept bids")

        start = self.clock.cycle(when=reservation.get_requested_term().get_new_start_time())

        start -= self.ADVANCE_TIME

        intervals = int((start - self.last_allocation)/self.CALL_INTERVAL)

        if intervals <= 0:
            intervals = 1

        start = self.last_allocation + (intervals * self.CALL_INTERVAL) + self.ADVANCE_TIME

        return start

    def get_approved_term(self, *, reservation: IBrokerReservation) -> Term:
        return Term(start=reservation.get_requested_term().get_start_time(),
                    end=reservation.get_requested_term().get_end_time(),
                    new_start=reservation.get_requested_term().get_new_start_time())

    def get_next_allocation(self, *, cycle: int) -> int:
        return self.last_allocation + self.CALL_INTERVAL

    def get_renew(self, *, reservation: IClientReservation) -> int:
        new_start_cycle = self.actor.get_actor_clock().cycle(when=reservation.get_term().get_end_time()) + 1
        return new_start_cycle - self.ADVANCE_TIME - self.CLOCK_SKEW

    def get_start_for_allocation(self, *, allocation_cycle: int) -> int:
        return allocation_cycle + self.ADVANCE_TIME

    def get_end_for_allocation(self, *, allocation_cycle: int) -> int:
        return allocation_cycle + self.ADVANCE_TIME + self.allocation_horizon

    def prepare(self, *, cycle: int):
        if not self.ready:
            self.last_allocation = cycle - self.CALL_INTERVAL
            self.ready = True

        try:
            self.check_pending()
        except Exception as e:
            self.logger.error("Exception in prepare {}".format(e))

    def process_renewing(self, *, renewing: ReservationSet, pending: ReservationSet) -> ReservationSet:
        """
        Performs checks on renewing reservations. Updates the terms to
        suggest new terms, stores the extend on the pending list. Returns a
        fresh ReservationSet of expiring reservations to try to renew in this
        bidding cycle.

        @param renewing collection of the renewing reservations
        @param pending collection of reservations that are pending

        @return non-null set of renewals
        """
        result = ReservationSet()
        if renewing is None:
            return None

        #self.logger.debug("Expiring = {}".format(renewing.size()))

        for reservation in renewing.values():
            self.logger.debug("Expiring res: {}".format(reservation))

            if reservation.is_renewable():
                self.logger.debug("This is a renewable expiring reservtion")

                term = reservation.get_term()

                term = term.extend()

                reservation.set_approved(term=term, approved_resources=reservation.get_resources().abstract_clone())

                result.add(reservation=reservation)
                self.calendar.add_pending(reservation=reservation)
            else:
                self.logger.debug("This is not a renewable expiring res")

        return result

    def query(self, *, p: dict) -> dict:
        """
        Returns the Broker Query Model
        @params p : dictionary containing filters (not used currently)
        """
        result = {}
        self.logger.debug("Processing Query with properties: {}".format(p))

        try:
            self.lock.acquire()
            if self.combined_broker_model is not None:
                graph = self.combined_broker_model.get_bqm(some=5)
                result[Constants.broker_query_model] = graph.serialize_graph()
                graph.delete_graph()
        finally:
            self.lock.release()

        self.logger.debug("Returning Query Result: {}".format(result))
        return result

    def satisfy_allocation(self, *, reservation: IBrokerReservation, source: IClientReservation, resource_share: int,
                           start_res_time: int):
        """
        These are request upcalls from the mapper with the manager lock held.
        Each request is a "bid". Here we simply validate the bids and save them
        in various calendar structures, pending action by an allocation policy
        that runs periodically ("auction").

        Performs all of the checks and necessary functions to allocate a
        ticket to a reservation from a source. Performs time shifting of a
        request's term if allowed, extracts units from the source, and assigns
        reservation.

        @param reservation request being allocated resources
        @param source source request is being allocated from
        @param resource_share number of resources being allocated to the request
        @param start_res_time when the ticket for the request starts

        @throws Exception in case of error
        """
        approved = self.get_approved_term(reservation=reservation)

        # Shift the requested term to start at the start time
        if self.clock.cycle(when=approved.get_new_start_time()) != start_res_time:
            if PropList.is_elastic_time(rset=reservation.get_requested_resources()):
                self.logger.debug("Reservation {} was schedules to start at {} is being shifted to {}"
                                  .format(reservation, self.clock.cycle(when=approved.get_new_start_time()),
                                          start_res_time))
                approved = approved.shift(date=self.clock.cycle_start_date(cycle=start_res_time))
            else:
                reservation.fail(message="Reservation has a different start time and time shifting is not allowed",
                                 exception=None)

        # Make sure the term end is aligned on a cycle boundary. NOTE: this
        # should also be aligned on call interval boundary!

        # calculate the cycle
        cycle = self.clock.cycle(when=approved.get_end_time())

        end_millis = self.clock.cycle_end_in_millis(cycle)

        # get the date that represents the end of the cycle
        aligned_end = ActorClock.from_milliseconds(milli_seconds=end_millis)
        # update the term end
        approved.set_end_time(date=aligned_end)

        # Get the ticket with the correct resources
        self.extract_ticket(reservation=reservation, source=source, approved=approved,
                            resource_share=resource_share)

        # Add to the calendar
        self.add_to_calendar(reservation=reservation)

        # Whatever happened up there, this bid is no longer pending. It either
        # succeeded or is now marked failed. (Could/should assert.)
        reservation.set_bid_pending(value=False)

    def verify_wanted_resources(self, *, reservation: IBrokerReservation, source: IClientReservation, start_time: int,
                                wanted: int):
        """
        Ensures that there are sufficient resources to allocate the
        number of requested resources.

        @param reservation reservation being allocated
        @param source source proposed for the requesting reservation
        @param start_time start of the proposed ticket
        @param wanted number of resources being requested

        @return true if there are sufficient resources,
                false otherwise

        @throws Exception in case of error
        """
        # calculate the number of available units
        available = self.calculate_available(start_time=start_time, source=source)
        mywanted = wanted

        self.logger.debug("ASPPlugin:allocateBids: allocating source rid({}) to newDynamicRes({}) that has props:{}"
                          .format(source.get_reservation_id(),
                                  reservation.get_reservation_id(),
                                  reservation.get_requested_resources().get_request_properties()))

        # Check if we have sufficient resources. If the reservation is flexible
        # adjust its resource demand to what is available.
        if wanted > available:
            if PropList.is_elastic_size(rset=reservation.get_requested_resources()) and available > 0:
                self.logger.debug("broker is shrinking an elastic request")
                mywanted = available
        else:
            reservation.fail_warn(message="Selected source holding has insufficient resources: source {}"
                                          " available {}; request {}".format(source, available, reservation))

        # If the (possibly adjusted) demand fits in what we've got, then try to
        # allocate.
        if mywanted <= available:
            return True
        else:
            return False

    def donate_delegation(self, *, delegation: IDelegation):
        self.logger.debug("Donate Delegation")
        self.bind_delegation(delegation=delegation)
        try:
            self.lock.acquire()
            if delegation.get_delegation_id() in self.delegations:
                self.combined_broker_model.merge_adm(adm=delegation.get_graph())
                self.combined_broker_model.validate_graph()
                self.logger.debug("Donated Delegation: self.combined_broker_model: {}".format(
                    self.combined_broker_model.serialize_graph()))
            else:
                self.logger.debug("Delegation ignored")
        finally:
            self.lock.release()

    def closed_delegation(self, *, delegation: IDelegation):
        self.logger.debug("Close Delegation")
        # TODO remove the delegation from the combined broker model



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import threading
from datetime import datetime
from typing import TYPE_CHECKING

from fabric.actor.core.common.exceptions import BrokerException
from fabric.actor.core.policy.broker_simple_policy import BrokerSimplePolicy
from fabric.actor.core.policy.fifo_queue import FIFOQueue
from fabric.actor.core.time.actor_clock import ActorClock
from fabric.actor.core.util.reservation_set import ReservationSet
from fabric.actor.core.util.resource_type import ResourceType

if TYPE_CHECKING:
    from fabric.actor.core.apis.i_broker import IBroker
    from fabric.actor.core.apis.i_broker_reservation import IBrokerReservation


class BrokerPriorityPolicy(BrokerSimplePolicy):
    """
    BrokerPriorityPolicy allocates requests based on requestType
    priorities set in the configuration at the broker. There may be multiple
    requestTypes with the same priority. Within each priority class, requests are
    allocated FIFO. Within each priority class the policy gives priority to
    extending reservations followed by new reservations.
    """
    PropertyRequestTypeCount = "requestType.count"
    PropertyRequestTypeName = "requestType.name"
    PropertyRequestTypePriority = "requestType.priority"
    PropertyQueueType = "queue.type"

    QueueTypeNone = "none"
    QueueTypeFifo = "fifo"

    QueueThreshold = "queueThreshold"
    # Pool the client requested its resources to be allocated from.
    PropertyPoolId = "pool.id"

    def __init__(self, *, actor: IBroker):
        super().__init__(actor=actor)
        self.queue = None

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['logger']
        del state['actor']
        del state['clock']
        del state['initialized']

        del state['for_approval']
        del state['lock']

        del state['calendar']

        del state['last_allocation']
        del state['allocation_horizon']
        del state['ready']
        del state['queue']

        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.logger = None
        self.actor = None
        self.clock = None
        self.initialized = False

        self.lock = threading.Lock()
        self.calendar = None

        if self.required_approval:
            self.for_approval = ReservationSet()

        self.last_allocation = -1
        self.allocation_horizon = 0
        self.ready = False

        self.queue = None

    def configure(self, *, properties: dict):
        """
        Processes a list of configuration properties
        @param properties properties
        @throws Exception in case of error
        """
        super().configure(properties=properties)

        if self.PropertyQueueType in properties:
            queue_type = properties[self.PropertyQueueType]
            if queue_type == self.QueueTypeFifo:
                self.queue = FIFOQueue()
            elif queue_type == self.QueueTypeNone:
                self.logger.debug("No queue")
            else:
                raise BrokerException("Unsupported queue type: {}".format(queue_type))

    def align_end(self, *, when: datetime) -> datetime:
        """
        Aligns the specified date with the end of the closest cycle.

        @param when when to align

        @return date aligned with the end of the closes cycle
        """
        cycle = self.clock.cycle(when=when)
        time = self.clock.cycle_end_in_millis(cycle=cycle)
        return ActorClock.from_milliseconds(milli_seconds=time)

    def align_start(self, *, when: datetime) -> datetime:
        """
        Aligns the specified date with the start of the closest cycle.

        @param when when to align

        @return date aligned with the start of the closes cycle
        """
        cycle = self.clock.cycle(when=when)
        time = self.clock.cycle_start_in_millis(cycle=cycle)
        return ActorClock.from_milliseconds(milli_seconds=time)

    def get_current_pool_id(self, *, reservation: IBrokerReservation) -> ResourceType:
        return reservation.get_source().get_type()

    def get_requested_pool_id(self, *, reservation: IBrokerReservation):
        if reservation.get_requested_resources() is not None and\
                reservation.get_requested_resources().get_request_properties() is not None and\
                self.PropertyPoolId in reservation.get_requested_resources().get_request_properties():
            return reservation.get_requested_resources().get_request_properties()[self.PropertyPoolId]

        return None



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations
from typing import TYPE_CHECKING, List

from fabric.actor.core.common.exceptions import ManageException
from fabric.actor.core.manage.authority_management_object import AuthorityManagementObject
from fabric.actor.core.apis.i_mgmt_authority import IMgmtAuthority
from fabric.actor.core.manage.local.local_server_actor import LocalServerActor
from fabric.message_bus.messages.reservation_mng import ReservationMng

if TYPE_CHECKING:
    from fabric.actor.core.manage.management_object import ManagementObject
    from fabric.actor.security.auth_token import AuthToken
    from fabric.actor.core.util.id import ID
    from fabric.message_bus.messages.unit_avro import UnitAvro
    from fabric.actor.core.apis.i_mgmt_actor import IMgmtActor


class LocalAuthority(LocalServerActor, IMgmtAuthority):
    def __init__(self, *, manager: ManagementObject, auth: AuthToken):
        super().__init__(manager=manager, auth=auth)

        if not isinstance(manager, AuthorityManagementObject):
            raise ManageException("Invalid manager object. Required: {}".format(type(AuthorityManagementObject)))

    def get_authority_reservations(self, *, id_token: str = None) -> List[ReservationMng]:
        self.clear_last()
        try:
            result = self.manager.get_authority_reservations(caller=self.auth)
            self.last_status = result.status
            if result.status.get_code() == 0:
                return result.result
        except Exception as e:
            self.last_exception = e

        return None

    def get_reservation_units(self, *, rid: ID, id_token: str = None) -> List[UnitAvro]:
        self.clear_last()
        try:
            result = self.manager.get_reservation_units(caller=self.auth, rid=rid)
            self.last_status = result.status
            if result.status.get_code() == 0:
                return result.result
        except Exception as e:
            self.last_exception = e

        return None

    def get_reservation_unit(self, *, uid: ID, id_token: str = None) -> UnitAvro:
        self.clear_last()
        try:
            result = self.manager.get_reservation_unit(caller=self.auth, uid=uid)
            self.last_status = result.status
            if result.get_status().get_code() == 0:
                return self.get_first(result_list=result.result)
        except Exception as e:
            self.last_exception = e

        return None

    def clone(self) -> IMgmtActor:
        return LocalAuthority(manager=self.manager, auth=self.auth)